/* $Id: cbench.c 91 2004-09-16 15:02:07Z /CN=thome/emailAddress=thome@loria.fr $ */

#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>

#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include "mploc.h"
#include "zp.h"
#include "z2.h"
#include "z2x.h"
#include "gmp_hacks.h"
#include "manu.h"
#include "timings.h"

#define	MPLOC_BANNER	"\n\n"\
"########################################"	\
"########################################\n"	\
"####\n"					\
"#### Mploc\n"					\
"####\n"


/* Zp */

/* Correctness */

#define	ZP_DEBUG_PRINT(v,n)		\
	BEGIN_BLOCK			\
		printf("%s:=",n);	\
		mploc_zp_print(v);	\
		printf("\n");		\
	END_BLOCK

#define CHECKTEST(msg,p) /* moan if a!=b, or precision less than p. */	\
	BEGIN_BLOCK							\
		if (a->prec<p || b->prec <p || !mploc_zp_eq(a,b)){	\
			ok=0;						\
			fprintf(stderr,"FAILED TEST [" msg "]\n");	\
			ZP_DEBUG_PRINT(x,"x");				\
			ZP_DEBUG_PRINT(y,"y");				\
			ZP_DEBUG_PRINT(z,"z");				\
			ZP_DEBUG_PRINT(a,"a");				\
			ZP_DEBUG_PRINT(b,"b");				\
			ZP_DEBUG_PRINT(c,"c");				\
			exit(255);					\
		}							\
	END_BLOCK
		
#undef	CORRECTNESS_NTESTS
#undef	RANDOM_ELEMENT_PRC
#undef	CORRECTNESS_PRC

#define CORRECTNESS_NTESTS	100
#define	RANDOM_ELEMENT_PRC	10
#define	CORRECTNESS_PRC		1000

mpz_t seed;

#define	gmp_getseed(s,prng)	mpz_set(seed,prng->_mp_seed)

mp_limb_t old_random_ui(gmp_randstate_t prng, mp_limb_t m)
{
	mpz_t x;
	mp_limb_t r;
	if (m==0) return m;	// otherwise mpz_urandomm crashes
	mpz_init_set_ui(x,m);
	mpz_urandomm(x,prng,x);
	r=mpz_get_ui(x);
	mpz_clear(x);
	return r;
}

void	correctness_zp(mplf_t zp)
{
	mploc_t x,y,z,a,b,c;
	gmp_randstate_t prng;
	int fl,ok,i;

	/* Elements */
	mploc_zp_init(x,zp);
	mploc_zp_init(y,zp);
	mploc_zp_init(z,zp);
	mploc_zp_init(a,zp);
	mploc_zp_init(b,zp);
	mploc_zp_init(c,zp);

	gmp_randinit_default(prng);
	
	/***********************************************************/
	printf("\nTesting addition...");
	ok=1;
	
	push_time();
	
	for(i=0;i<CORRECTNESS_NTESTS;i++) {
		gmp_getseed(seed,prng);
		gmp_randseed(prng,seed);

		mploc_zp_set_random(x,prng,RANDOM_ELEMENT_PRC);
		mploc_zp_set_random(y,prng,RANDOM_ELEMENT_PRC);
		mploc_zp_set_random(z,prng,RANDOM_ELEMENT_PRC);

		mploc_zp_force_precision(x,CORRECTNESS_PRC);
		mploc_zp_force_precision(y,CORRECTNESS_PRC);
		mploc_zp_force_precision(z,CORRECTNESS_PRC);

		mploc_zp_add(a,x,y);
		mploc_zp_add(b,a,z);
		mploc_zp_add(c,y,z);
		mploc_zp_add(a,x,c);
		CHECKTEST("addition associativity",CORRECTNESS_PRC);

		mploc_zp_add(a,x,y);
		mploc_zp_add(a,a,z);
		mploc_zp_add(b,y,z);
		mploc_zp_add(b,b,x);
		CHECKTEST("add. assoc. w/ left overlap",CORRECTNESS_PRC);

		mploc_zp_add(a,x,y);
		mploc_zp_add(a,z,a);
		mploc_zp_add(b,y,z);
		mploc_zp_add(b,x,b);
		CHECKTEST("add. assoc. w/ right overlap",CORRECTNESS_PRC);

		mploc_zp_add(a,y,x);
		mploc_zp_add(b,x,y);
		CHECKTEST("add. commutativity",CORRECTNESS_PRC);
	}

	printf("%d tests %s\n", CORRECTNESS_NTESTS,
			ok?"passed":"done ; failures encountered");

	pop_time("Time for addition tests");

	/***********************************************************/
	printf("\nTesting multiplication...");
	ok=1;
	push_time();
	for(i=0;i<CORRECTNESS_NTESTS;i++) {
		gmp_getseed(seed,prng);
		gmp_randseed(prng,seed);

		mploc_zp_set_random(x,prng,RANDOM_ELEMENT_PRC);
		mploc_zp_set_random(y,prng,RANDOM_ELEMENT_PRC);
		mploc_zp_set_random(z,prng,RANDOM_ELEMENT_PRC);

		mploc_zp_force_precision(x,CORRECTNESS_PRC);
		mploc_zp_force_precision(y,CORRECTNESS_PRC);
		mploc_zp_force_precision(z,CORRECTNESS_PRC);

		mploc_zp_mul(a,x,y);
		mploc_zp_mul(b,a,z);
		mploc_zp_mul(c,y,z);
		mploc_zp_mul(a,x,c);
		CHECKTEST("mult. associativity",CORRECTNESS_PRC);

		mploc_zp_mul(a,x,y);
		mploc_zp_mul(a,a,z);
		mploc_zp_mul(b,y,z);
		mploc_zp_mul(b,b,x);
		CHECKTEST("mult. assoc. w/ left overlap",CORRECTNESS_PRC);

		mploc_zp_mul(a,x,y);
		mploc_zp_mul(a,z,a);
		mploc_zp_mul(b,y,z);
		mploc_zp_mul(b,x,b);
		CHECKTEST("mult. assoc. w/ right overlap",CORRECTNESS_PRC);

		mploc_zp_mul(a,y,x);
		mploc_zp_mul(b,x,y);
		CHECKTEST("mult. commutativity",CORRECTNESS_PRC);

		mploc_zp_mul(b,z,x);
		mploc_zp_mul(c,z,y);
		mploc_zp_add(a,b,c);
		mploc_zp_add(c,x,y);
		mploc_zp_mul(b,z,c);
		CHECKTEST("mult. left distributivity",CORRECTNESS_PRC);
	}

	printf("%d tests %s\n", CORRECTNESS_NTESTS,
			ok?"passed":"done ; failures encountered");

	pop_time("Time for multiplication tests");

	/***********************************************************/
	printf("\nTesting inversion...");
	ok=1; fl=0;
	
	push_time();
	for(i=0;i<CORRECTNESS_NTESTS;i++) {

		gmp_getseed(seed,prng);
		gmp_randseed(prng,seed);
		mploc_zp_set_random(x,prng,RANDOM_ELEMENT_PRC);
		mploc_zp_force_precision(x,CORRECTNESS_PRC);

		if (mploc_zp_valuation(x)>0) {
			fl++;
			continue;
		}

		mploc_zp_inv(y,x);
		mploc_zp_add_ui(y,y,1);

		if (mploc_zp_valuation(y)>0) {
			fl++;
			continue;
		}

		mploc_zp_inv(z,y);
		mploc_zp_add_ui(c,x,1);
		mploc_zp_mul(b,c,z);
		mploc_zp_set(a,x);
		CHECKTEST("inversion",CORRECTNESS_PRC);

		/* Contents of registers after test:
			x = x
			y = 1+1/x
			z = 1/z = x/(x+1)
			a = x
			b = (x+1)*x/(x+1)
			c = x+1
		*/
	}

	printf("%d tests %s ; %d bad inputs\n", CORRECTNESS_NTESTS,
			ok?"passed":"done ; failures encountered",
			fl);

	pop_time("Time for inversion tests");
	
	/***********************************************************/
	printf("\nTesting invsqrt...");
	ok=1;
	fl=0;
			
	push_time();

	for(i=0;i<2*CORRECTNESS_NTESTS;i++) {
		gmp_getseed(seed,prng);
		gmp_randseed(prng,seed);
		mploc_zp_set_random(x,prng,RANDOM_ELEMENT_PRC);
		mploc_zp_force_precision(x,CORRECTNESS_PRC);

		if (mploc_zp_valuation(x)>0) {
			fl++;
			continue;
		}

		if (!mploc_zp_invsqrt(y,x)) {
			fl++;
			continue;
		}

		mploc_zp_mul(z,y,y);
		mploc_zp_mul(a,x,z);
		mploc_zp_set_ui(b,1,CORRECTNESS_PRC);
		CHECKTEST("invsqrt",CORRECTNESS_PRC);

		/* Contents of registers after test:
			x = x
			y = 1/sqrt(x)
			z = 1/x
			a = x/x
			b = 1
		*/
	}
	printf("%d tests %s ; %d bad inputs\n", CORRECTNESS_NTESTS,
			ok?"passed":"done ; failures encountered",
			fl);
	pop_time("Time for invsqrt tests");

	gmp_randclear(prng);

	mploc_zp_clear(x);
	mploc_zp_clear(y);
	mploc_zp_clear(z);
	mploc_zp_clear(a);
	mploc_zp_clear(b);
	mploc_zp_clear(c);
}
#undef CHECKTEST


extern void mploc_zp_inversion_configure(int, int);

static int precisions_to_test[] UNUSED_VARIABLE ={
	1,2,
	3,4,5,6,
	7,8,9,12,
	15,16,17,24,
	31,32,33,48,
	63,64,65,96,
	127,128,129,192,
	255,256,257,384,
	511,512,513};

static int seeds_to_test[] UNUSED_VARIABLE ={
	1,2,
	3,4,6,
	7,8,12,
	15,16,24,
	31,32,48,
	63,64,96,
	127,128,192,
	255,256,384,
	511,512};

#define	NPRC	(sizeof(precisions_to_test)/sizeof(mp_size_t))
#define	NSEEDS	(sizeof(seeds_to_test)/sizeof(int))
#define	FIXED_WIDTH_PRC	100
#define	TIMINGS_NTESTS	100

void	fixed_width_timings_zp(mplf_t zp)
{
	int	i;
	mploc_t x,y,z,a,b,c;
	gmp_randstate_t prng;
	
	mploc_zp_init(x,zp);
	mploc_zp_init(y,zp);
	mploc_zp_init(z,zp);
	mploc_zp_init(a,zp);
	mploc_zp_init(b,zp);
	mploc_zp_init(c,zp);

	gmp_randinit_default(prng);
	
	timings_set_dispmode(TIMINGS_ARE_HUMAN_READABLE);

	printf("\nFixed-width timings (precision=%d)\n",FIXED_WIDTH_PRC);

	/**********************************************************/
	mploc_zp_set_random(x,prng,FIXED_WIDTH_PRC);
	mploc_zp_set_random(y,prng,FIXED_WIDTH_PRC);

	push_time();
	for(i=0;i<TIMINGS_NTESTS;i++) {
		mploc_zp_add(z,x,y);
	}
	pop_time("fixed-width addition");

	/**********************************************************/
	mploc_zp_set_random(x,prng,FIXED_WIDTH_PRC);
	mploc_zp_set_random(y,prng,FIXED_WIDTH_PRC);

	push_time();
	for(i=0;i<TIMINGS_NTESTS;i++) {
		mploc_zp_mul(z,x,y);
	}
	pop_time("fixed-width multiplication");

	/**********************************************************/
	mploc_zp_set_random(x,prng,FIXED_WIDTH_PRC);

	if (mploc_zp_valuation(x)>0) {
		mploc_zp_add_ui(x,x,1);
	}

	mploc_zp_inversion_configure(1,0);
	
	push_time();
	for(i=0;i<TIMINGS_NTESTS;i++) {
		mploc_zp_inv(z,x);
	}
	pop_time("fixed-width inversion");

	/**********************************************************/
	mploc_zp_set_random(x,prng,FIXED_WIDTH_PRC);

	if (mploc_zp_valuation(x)>0) {
		mploc_zp_add_ui(x,x,1);
		mploc_zp_mul(x,x,x);
	}

	push_time();
	for(i=0;i<TIMINGS_NTESTS;i++) {
		mploc_zp_invsqrt(z,x);
	}
	pop_time("fixed-width invsqrt");

	gmp_randclear(prng);

	mploc_zp_clear(x);
	mploc_zp_clear(y);
	mploc_zp_clear(z);
	mploc_zp_clear(a);
	mploc_zp_clear(b);
	mploc_zp_clear(c);
}

void	variable_width_timings_zp(mplf_t zp)
{
	int	i;
	int	j;
	mploc_t x,y,z,a,b,c;
	mp_size_t curr_prec;
	gmp_randstate_t prng;
	
	mploc_zp_init(x,zp);
	mploc_zp_init(y,zp);
	mploc_zp_init(z,zp);
	mploc_zp_init(a,zp);
	mploc_zp_init(b,zp);
	mploc_zp_init(c,zp);

	gmp_randinit_default(prng);
	
	timings_set_dispmode(TIMINGS_ARE_SECONDS);

	printf("\nVariable-width timings\n");

	/**********************************************************/
	fprintf(stderr,"Benchmarking multiplication\n");

	for(j=0;j<NPRC;j++) {
		curr_prec = precisions_to_test[j];

		mploc_zp_set_random(x,prng,curr_prec);
		mploc_zp_set_random(y,prng,curr_prec);

		push_time();
		for(i=0;i<TIMINGS_NTESTS;i++) {
			mploc_zp_mul(x,x,y);
			mploc_zp_normalize(x);
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}

	/**********************************************************/
	fprintf(stderr,"Benchmarking full-extent inversion (iterative)\n");
	mploc_zp_inversion_configure(0,0);

	for(j=1;j<NPRC;j++) {
		curr_prec = precisions_to_test[j];

		do {
			mploc_zp_set_random(x,prng,curr_prec);
		} while (mploc_zp_valuation(x)>0);

		push_time();
		for(i=0;i<TIMINGS_NTESTS;i++) {
			mploc_zp_inv(y,x);
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}

	/**********************************************************/
	fprintf(stderr,"Benchmarking full-extent inversion (recursive)\n");
	mploc_zp_inversion_configure(1,0);

	for(j=1;j<NPRC;j++) {
		curr_prec = precisions_to_test[j];

		do {
			mploc_zp_set_random(x,prng,curr_prec);
		} while (mploc_zp_valuation(x)>0);

		push_time();
		for(i=0;i<TIMINGS_NTESTS;i++) {
			mploc_zp_inv(y,x);
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}

	/**********************************************************/
	fprintf(stderr,"Benchmarking seeded inversion (recursive)\n");

	for(j=0;j<NSEEDS;j++) {
		mploc_zp_inversion_configure(1,seeds_to_test[j]);
		push_time();

		for(i=0;i<TIMINGS_NTESTS;i++) {
			mploc_zp_set_random(x,prng,FIXED_WIDTH_PRC);
			if (mploc_zp_valuation(x)>0) continue;

			mploc_zp_inv(y,x);
		}
		printf("%d ",seeds_to_test[j]);
		pop_time_raw();
		printf("\n");
	}

	/**********************************************************/
	fprintf(stderr,"Benchmarking invsqrt\n");

	for(j=1;j<NPRC;j++) {
		curr_prec = precisions_to_test[j];

		do {
			mploc_zp_set_random(x,prng,curr_prec);
		} while (mploc_zp_valuation(x)>0);

		mploc_zp_mul(x,x,x);

		push_time();
		for(i=0;i<TIMINGS_NTESTS;i++) {
			mploc_zp_invsqrt(y,x);
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}

	gmp_randclear(prng);

	mploc_zp_clear(x);
	mploc_zp_clear(y);
	mploc_zp_clear(z);
	mploc_zp_clear(a);
	mploc_zp_clear(b);
	mploc_zp_clear(c);
}

/* Z2 */

/* Correctness */

#define	Z2_DEBUG_PRINT(v,n)		\
	BEGIN_BLOCK			\
		printf("%s:=",n);	\
		mploc_z2_print(v);	\
		printf("\n");		\
	END_BLOCK

#define CHECKTEST(msg,p) /* moan if a!=b, or precision less than p. */	\
	BEGIN_BLOCK							\
		if (a->prec<p || b->prec <p || !mploc_z2_eq(a,b)){	\
			ok=0;						\
			fprintf(stderr,"FAILED TEST [" msg "]\n");	\
			Z2_DEBUG_PRINT(x,"x");				\
			Z2_DEBUG_PRINT(y,"y");				\
			Z2_DEBUG_PRINT(z,"z");				\
			Z2_DEBUG_PRINT(a,"a");				\
			Z2_DEBUG_PRINT(b,"b");				\
			Z2_DEBUG_PRINT(c,"c");				\
			exit(255);					\
		}							\
	END_BLOCK
		
#undef	CORRECTNESS_NTESTS
#undef	RANDOM_ELEMENT_PRC
#undef	CORRECTNESS_PRC

#define CORRECTNESS_NTESTS	100
#define	RANDOM_ELEMENT_PRC	10
#define	CORRECTNESS_PRC		33

void	correctness_z2(mplf_t z2)
{
	mploc_t x,y,z,a,b,c;
	mpz_t	r;
	gmp_randstate_t prng;
	int fl,ok,i;

	/* Elements */
	mploc_z2_init(x,z2);
	mploc_z2_init(y,z2);
	mploc_z2_init(z,z2);
	mploc_z2_init(a,z2);
	mploc_z2_init(b,z2);
	mploc_z2_init(c,z2);

	gmp_randinit_default(prng);
	
	mpz_init(r);

	/***********************************************************/
	printf("\nTesting addition...");
	ok=1;
	
	push_time();
	
	for(i=0;i<CORRECTNESS_NTESTS;i++) {
		gmp_getseed(seed,prng);
		gmp_randseed(prng,seed);

		mploc_z2_set_random(x,prng,RANDOM_ELEMENT_PRC);
		mploc_z2_set_random(y,prng,RANDOM_ELEMENT_PRC);
		mploc_z2_set_random(z,prng,RANDOM_ELEMENT_PRC);
		mploc_z2_force_precision(x,CORRECTNESS_PRC);
		mploc_z2_force_precision(y,CORRECTNESS_PRC);
		mploc_z2_force_precision(z,CORRECTNESS_PRC);

		mploc_z2_add(a,x,y);
		mploc_z2_add(b,a,z);
		mploc_z2_add(c,y,z);
		mploc_z2_add(a,x,c);
		CHECKTEST("addition associativity",CORRECTNESS_PRC);

		mploc_z2_add(a,x,y);
		mploc_z2_add(a,a,z);
		mploc_z2_add(b,y,z);
		mploc_z2_add(b,b,x);
		CHECKTEST("add. assoc. w/ left overlap",CORRECTNESS_PRC);

		mploc_z2_add(a,x,y);
		mploc_z2_add(a,z,a);
		mploc_z2_add(b,y,z);
		mploc_z2_add(b,x,b);
		CHECKTEST("add. assoc. w/ right overlap",CORRECTNESS_PRC);

		mploc_z2_add(a,y,x);
		mploc_z2_add(b,x,y);
		CHECKTEST("add. commutativity",CORRECTNESS_PRC);
	}

	printf("%d tests %s\n", CORRECTNESS_NTESTS,
			ok?"passed":"done ; failures encountered");

	pop_time("Time for addition tests");

	/***********************************************************/
	printf("\nTesting multiplication...");
	ok=1;
	push_time();
	for(i=0;i<CORRECTNESS_NTESTS;i++) {
		gmp_getseed(seed,prng);
		gmp_randseed(prng,seed);

		mploc_z2_set_random(x,prng,RANDOM_ELEMENT_PRC);
		mploc_z2_set_random(y,prng,RANDOM_ELEMENT_PRC);
		mploc_z2_set_random(z,prng,RANDOM_ELEMENT_PRC);
		mploc_z2_force_precision(x,CORRECTNESS_PRC);
		mploc_z2_force_precision(y,CORRECTNESS_PRC);
		mploc_z2_force_precision(z,CORRECTNESS_PRC);

		mploc_z2_mul(a,x,y);
		mploc_z2_mul(b,a,z);
		mploc_z2_mul(c,y,z);
		mploc_z2_mul(a,x,c);
		CHECKTEST("mult. associativity",CORRECTNESS_PRC);

		mploc_z2_mul(a,x,y);
		mploc_z2_mul(a,a,z);
		mploc_z2_mul(b,y,z);
		mploc_z2_mul(b,b,x);
		CHECKTEST("mult. assoc. w/ left overlap",CORRECTNESS_PRC);

		mploc_z2_mul(a,x,y);
		mploc_z2_mul(a,z,a);
		mploc_z2_mul(b,y,z);
		mploc_z2_mul(b,x,b);
		CHECKTEST("mult. assoc. w/ right overlap",CORRECTNESS_PRC);

		mploc_z2_mul(a,y,x);
		mploc_z2_mul(b,x,y);
		CHECKTEST("mult. commutativity",CORRECTNESS_PRC);

		mploc_z2_mul(b,z,x);
		mploc_z2_mul(c,z,y);
		mploc_z2_add(a,b,c);
		mploc_z2_add(c,x,y);
		mploc_z2_mul(b,z,c);
		CHECKTEST("mult. left distributivity",CORRECTNESS_PRC);
	}

	printf("%d tests %s\n", CORRECTNESS_NTESTS,
			ok?"passed":"done ; failures encountered");

	pop_time("Time for multiplication tests");

	/***********************************************************/
	printf("\nTesting inversion...");
	ok=1;
	fl=0;
	
	push_time();
	for(i=0;i<CORRECTNESS_NTESTS;i++) {
		gmp_getseed(seed,prng);
		gmp_randseed(prng,seed);

		mploc_z2_set_random(x,prng,RANDOM_ELEMENT_PRC);
		mploc_z2_force_precision(x,CORRECTNESS_PRC);

		if (mploc_z2_valuation(x)>0) {
			fl++;
			continue;
		}

		mploc_z2_inv(y,x);
		mploc_z2_lshift(y,y,1);
		mploc_z2_add_ui(y,y,1);
		mploc_z2_inv(z,y);
		mploc_z2_add_ui(c,x,2);
		mploc_z2_mul(b,c,z);
		mploc_z2_set(a,x);
		CHECKTEST("inversion",CORRECTNESS_PRC);

		mploc_z2_inv(x,x);
		mploc_z2_inv(b,x);
		CHECKTEST("inversion w/ overlap",CORRECTNESS_PRC);

		/* Contents of registers after test:
			x = x
			y = 1+2/x
			z = 1/y = x/(x+2)
			a = x
			b = (x+1)*x/(x+1)
			c = x+1
		*/
	}

	printf("%d tests %s ; %d bad inputs\n", CORRECTNESS_NTESTS,
			ok?"passed":"done ; failures encountered",
			fl);

	pop_time("Time for inversion tests");
	
	/***********************************************************/
	printf("\nTesting invsqrt...");
	ok=1;
	fl=0;
			
	push_time();

	for(i=0;i<2*CORRECTNESS_NTESTS;i++) {
		gmp_getseed(seed,prng);
		gmp_randseed(prng,seed);
		mploc_z2_set_random(x,prng,RANDOM_ELEMENT_PRC);
		mploc_z2_force_precision(x,CORRECTNESS_PRC);


		if (mploc_z2_valuation(x)>0) {
			fl++;
			continue;
		}

		if (!mploc_z2_invsqrt(y,x)) {
			fl++;
			continue;
		}

		mploc_z2_mul(z,y,y);
		mploc_z2_mul(a,x,z);
		mploc_z2_set_ui(b,1,CORRECTNESS_PRC);
		CHECKTEST("invsqrt",CORRECTNESS_PRC-1);

		/* Contents of registers after test:
			x = x
			y = 1/sqrt(x)
			z = 1/x
			a = x/x
			b = 1
		*/
	}
	printf("%d tests %s ; %d bad inputs\n", CORRECTNESS_NTESTS,
			ok?"passed":"done ; failures encountered",
			fl);
	pop_time("Time for invsqrt tests");

	gmp_randclear(prng);

	mploc_z2_clear(x);
	mploc_z2_clear(y);
	mploc_z2_clear(z);
	mploc_z2_clear(a);
	mploc_z2_clear(b);
	mploc_z2_clear(c);
}
#undef CHECKTEST

extern void mploc_z2_inversion_configure(int, int);

void	fixed_width_timings_z2(mplf_t z2)
{
	int	i;
	mploc_t x,y,z,a,b,c;
	mpz_t	r;
	gmp_randstate_t prng;
	
	mploc_z2_init(x,z2);
	mploc_z2_init(y,z2);
	mploc_z2_init(z,z2);
	mploc_z2_init(a,z2);
	mploc_z2_init(b,z2);
	mploc_z2_init(c,z2);

	gmp_randinit_default(prng);
	
	mpz_init(r);

	timings_set_dispmode(TIMINGS_ARE_HUMAN_READABLE);

	printf("\nFixed-width timings (precision=%d)\n",FIXED_WIDTH_PRC);

	/**********************************************************/
	mploc_z2_set_random(x,prng,FIXED_WIDTH_PRC);
	mploc_z2_set_random(y,prng,FIXED_WIDTH_PRC);

	push_time();
	for(i=0;i<TIMINGS_NTESTS;i++) {
		mploc_z2_add(z,x,y);
	}
	pop_time("fixed-width addition");

	/**********************************************************/
	mploc_z2_set_random(x,prng,FIXED_WIDTH_PRC);
	mploc_z2_set_random(y,prng,FIXED_WIDTH_PRC);

	push_time();
	for(i=0;i<TIMINGS_NTESTS;i++) {
		mploc_z2_mul(z,x,y);
	}
	pop_time("fixed-width multiplication");

	/**********************************************************/
	mploc_z2_set_random(x,prng,FIXED_WIDTH_PRC);

	if (mploc_z2_valuation(x)>0) {
		mploc_z2_add_ui(x,x,1);
	}

	mploc_z2_inversion_configure(1,0);
	
	push_time();
	for(i=0;i<TIMINGS_NTESTS;i++) {
		mploc_z2_inv(z,x);
	}
	pop_time("fixed-width inversion");

	/**********************************************************/
	mploc_z2_set_random(x,prng,FIXED_WIDTH_PRC);

	if (mploc_z2_valuation(x)>0) {
		mploc_z2_add_ui(x,x,1);
		mploc_z2_mul(x,x,x);
	}

	push_time();
	for(i=0;i<TIMINGS_NTESTS;i++) {
		mploc_z2_invsqrt(z,x);
	}
	pop_time("fixed-width invsqrt");

	gmp_randclear(prng);

	mploc_z2_clear(x);
	mploc_z2_clear(y);
	mploc_z2_clear(z);
	mploc_z2_clear(a);
	mploc_z2_clear(b);
	mploc_z2_clear(c);
}

void	variable_width_timings_z2(mplf_t z2)
{
	int	i;
	int	j;
	mploc_t x,y,z,a,b,c;
	mp_size_t curr_prec;
	gmp_randstate_t prng;
	
	mploc_z2_init(x,z2);
	mploc_z2_init(y,z2);
	mploc_z2_init(z,z2);
	mploc_z2_init(a,z2);
	mploc_z2_init(b,z2);
	mploc_z2_init(c,z2);

	gmp_randinit_default(prng);
	
	timings_set_dispmode(TIMINGS_ARE_SECONDS);

	printf("\nVariable-width timings\n");

	/**********************************************************/
	fprintf(stderr,"Benchmarking multiplication\n");

	for(j=0;j<NPRC;j++) {
		curr_prec = precisions_to_test[j];

		mploc_z2_set_random(x,prng,curr_prec);
		mploc_z2_set_random(y,prng,curr_prec);

		push_time();
		for(i=0;i<TIMINGS_NTESTS;i++) {
			mploc_z2_mul(x,x,y);
			mploc_z2_normalize(x);
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}

	/**********************************************************/
	fprintf(stderr,"Benchmarking full-extent inversion (iterative)\n");
	mploc_z2_inversion_configure(0,0);

	for(j=1;j<NPRC;j++) {
		curr_prec = precisions_to_test[j];

		do {
			mploc_z2_set_random(x,prng,curr_prec);
		} while (mploc_z2_valuation(x)>0);

		push_time();
		for(i=0;i<TIMINGS_NTESTS;i++) {
			mploc_z2_inv(y,x);
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}

#if 0	/* recursion is not implemented yet */
	/**********************************************************/
	fprintf(stderr,"Benchmarking full-extent inversion (recursive)\n");
	mploc_z2_inversion_configure(1,0);

	for(j=1;j<NPRC;j++) {
		curr_prec = precisions_to_test[j];

		do {
			mploc_z2_set_random(x,prng,curr_prec);
		} while (mploc_z2_valuation(x)>0);

		push_time();
		for(i=0;i<TIMINGS_NTESTS;i++) {
			mploc_z2_inv(y,x);
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}

	/**********************************************************/
	fprintf(stderr,"Benchmarking seeded inversion (recursive)\n");

	for(j=0;j<NSEEDS;j++) {
		mploc_z2_inversion_configure(1,seeds_to_test[j]);

		do {
			mploc_z2_set_random(x,prng,FIXED_WIDTH_PRC);
		} while (mploc_z2_valuation(x)>0);

		push_time();
		for(i=0;i<TIMINGS_NTESTS;i++) {
			mploc_z2_inv(y,x);
		}
		printf("%d ",seeds_to_test[j]);
		pop_time_raw();
		printf("\n");
	}
#endif

	/**********************************************************/
	fprintf(stderr,"Benchmarking invsqrt\n");

	for(j=1;j<NPRC;j++) {
		curr_prec = precisions_to_test[j];

		do {
			mploc_z2_set_random(x,prng,curr_prec);
		} while (mploc_z2_valuation(x)>0);

		mploc_z2_mul(x,x,x);

		push_time();
		for(i=0;i<TIMINGS_NTESTS;i++) {
			mploc_z2_invsqrt(y,x);
		}
		printf("%lu ",curr_prec);
		pop_time_raw();
		printf("\n");
	}

	gmp_randclear(prng);

	mploc_z2_clear(x);
	mploc_z2_clear(y);
	mploc_z2_clear(z);
	mploc_z2_clear(a);
	mploc_z2_clear(b);
	mploc_z2_clear(c);
}

/* Z2X */

/* Correctness */

#define	Z2X_DEBUG_PRINT(v,n)		\
	BEGIN_BLOCK			\
		printf("%s:=",n);	\
		mploc_z2x_print(v);	\
		printf(";\n");		\
	END_BLOCK

#define CHECKTEST(msg,p) /* moan if a!=b, or precision less than p. */	\
	BEGIN_BLOCK							\
		if (a->prec<p || b->prec <p || !mploc_z2x_eq(a,b)){	\
			ok=0;						\
			fprintf(stderr,"FAILED TEST [" msg "]\n");	\
			Z2X_DEBUG_PRINT(x,"x");				\
			Z2X_DEBUG_PRINT(y,"y");				\
			Z2X_DEBUG_PRINT(z,"z");				\
			Z2X_DEBUG_PRINT(a,"a");				\
			Z2X_DEBUG_PRINT(b,"b");				\
			Z2X_DEBUG_PRINT(c,"c");				\
			exit(255);					\
		}							\
	END_BLOCK
		
#undef	CORRECTNESS_NTESTS
#undef	RANDOM_ELEMENT_PRC
#undef	CORRECTNESS_PRC
#undef	TIMINGS_NTESTS
#undef	FIXED_WIDTH_PRC

#define CORRECTNESS_NTESTS	1000
#define	RANDOM_ELEMENT_PRC	100
#define	CORRECTNESS_PRC		100
#define	CORRECTNESS_Z2X_DEGREE	100
#define	TIMINGS_NTESTS		100
#define	FIXED_WIDTH_PRC		120

// extern void set_kara_threshold(int);

void	correctness_z2x(mplf_t z2)
{
	mploc_z2x_t x,y,z,a,b,c;
	mpz_t	r;
	gmp_randstate_t prng;
	int fl,ok,i;
	mploc_t	scalar_elt;

	/* Elements */
	mploc_z2x_init(x,z2);
	mploc_z2x_init(y,z2);
	mploc_z2x_init(z,z2);
	mploc_z2x_init(a,z2);
	mploc_z2x_init(b,z2);
	mploc_z2x_init(c,z2);
	mploc_z2_init(scalar_elt,z2);

	gmp_randinit_default(prng);
	
	mpz_init(r);

	/***********************************************************/
	printf("\nTesting addition...");
	ok=1;
	
	push_time();
	
	for(i=0;i<CORRECTNESS_NTESTS;i++) {

		gmp_getseed(seed,prng);
		gmp_randseed(prng,seed);

		mploc_z2x_set_random(x,prng,old_random_ui(prng,CORRECTNESS_Z2X_DEGREE),RANDOM_ELEMENT_PRC);
		mploc_z2x_set_random(y,prng,old_random_ui(prng,CORRECTNESS_Z2X_DEGREE),RANDOM_ELEMENT_PRC);
		mploc_z2x_set_random(z,prng,old_random_ui(prng,CORRECTNESS_Z2X_DEGREE),RANDOM_ELEMENT_PRC);
		mploc_z2x_force_precision(x,CORRECTNESS_PRC);
		mploc_z2x_force_precision(y,CORRECTNESS_PRC);
		mploc_z2x_force_precision(z,CORRECTNESS_PRC);

		mploc_z2x_add(a,x,y);
		mploc_z2x_add(b,a,z);
		mploc_z2x_add(c,y,z);
		mploc_z2x_add(a,x,c);
		CHECKTEST("addition associativity",CORRECTNESS_PRC);

		mploc_z2x_add(a,x,y);
		mploc_z2x_add(a,a,z);
		mploc_z2x_add(b,y,z);
		mploc_z2x_add(b,b,x);
		CHECKTEST("add. assoc. w/ left overlap",CORRECTNESS_PRC);

		mploc_z2x_add(a,x,y);
		mploc_z2x_add(a,z,a);
		mploc_z2x_add(b,y,z);
		mploc_z2x_add(b,x,b);
		CHECKTEST("add. assoc. w/ right overlap",CORRECTNESS_PRC);

		mploc_z2x_add(a,y,x);
		mploc_z2x_add(b,x,y);
		CHECKTEST("add. commutativity",CORRECTNESS_PRC);
	}

	printf("%d tests %s\n", CORRECTNESS_NTESTS,
			ok?"passed":"done ; failures encountered");

	pop_time("Time for addition tests");


	/***********************************************************/
	printf("\nTesting multiplication...");
	ok=1;
	push_time();
	for(i=0;i<CORRECTNESS_NTESTS;i++) {

		/* This is to enable a quick recovery of the failure
		 * condition */
		gmp_getseed(seed,prng);
		gmp_randseed(prng,seed);

		mploc_z2x_set_random(x,prng,old_random_ui(prng,CORRECTNESS_Z2X_DEGREE),CORRECTNESS_PRC);
		mploc_z2x_set_random(y,prng,old_random_ui(prng,CORRECTNESS_Z2X_DEGREE),CORRECTNESS_PRC);
		mploc_z2x_set_random(z,prng,old_random_ui(prng,CORRECTNESS_Z2X_DEGREE),CORRECTNESS_PRC);
		mploc_z2x_force_precision(x,CORRECTNESS_PRC);
		mploc_z2x_force_precision(y,CORRECTNESS_PRC);
		mploc_z2x_force_precision(z,CORRECTNESS_PRC);

		mploc_z2x_mul(a,x,y);
		mploc_z2x_mul(b,a,z);
		mploc_z2x_mul(c,y,z);
		mploc_z2x_mul(a,x,c);
		CHECKTEST("mult. associativity",CORRECTNESS_PRC);

		mploc_z2x_mul(a,x,y);
		mploc_z2x_mul(a,a,z);
		mploc_z2x_mul(b,y,z);
		mploc_z2x_mul(b,b,x);
		CHECKTEST("mult. assoc. w/ left overlap",CORRECTNESS_PRC);

		mploc_z2x_mul(a,x,y);
		mploc_z2x_mul(a,z,a);
		mploc_z2x_mul(b,y,z);
		mploc_z2x_mul(b,x,b);
		CHECKTEST("mult. assoc. w/ right overlap",CORRECTNESS_PRC);

		mploc_z2x_mul(a,y,x);
		mploc_z2x_mul(b,x,y);
		CHECKTEST("mult. commutativity",CORRECTNESS_PRC);

		mploc_z2x_mul(b,z,x);
		mploc_z2x_mul(c,z,y);
		mploc_z2x_add(a,b,c);
		mploc_z2x_add(c,x,y);
		mploc_z2x_mul(b,z,c);
		CHECKTEST("mult. left distributivity",CORRECTNESS_PRC);
	}

	printf("%d tests %s\n", CORRECTNESS_NTESTS,
			ok?"passed":"done ; failures encountered");

	pop_time("Time for multiplication tests");

	/***********************************************************/
	printf("\nTesting division...");
	ok=1;fl=0;
	push_time();
	for(i=0;i<CORRECTNESS_NTESTS;i++) {

		gmp_getseed(seed,prng);
		gmp_randseed(prng,seed);

		mploc_z2x_set_random(x,prng,old_random_ui(prng,CORRECTNESS_Z2X_DEGREE),CORRECTNESS_PRC);
		mploc_z2x_set_random(y,prng,old_random_ui(prng,CORRECTNESS_Z2X_DEGREE),CORRECTNESS_PRC);
		mploc_z2x_set_random(z,prng,old_random_ui(prng,CORRECTNESS_Z2X_DEGREE),CORRECTNESS_PRC);
		mploc_z2x_force_precision(x,CORRECTNESS_PRC);
		mploc_z2x_force_precision(y,CORRECTNESS_PRC);
		mploc_z2x_force_precision(z,CORRECTNESS_PRC);

		if (y->deg==-1) {
			fl++;
			continue;
		}

		mploc_z2x_get_coeff(scalar_elt,y,y->deg);

		if (mploc_z2_valuation(scalar_elt)) {
			fl++;
			continue;
		}

		mploc_z2x_set(a,x);
		mploc_z2x_div(b,c,a,y);
		mploc_z2x_mul(b,b,y);
		mploc_z2x_add(b,b,c);
		CHECKTEST("division",CORRECTNESS_PRC);

		mploc_z2x_set(a,x);
		mploc_z2x_div(x,c,a,y);
		mploc_z2x_set(b,x);
		mploc_z2x_mul(b,b,y);
		mploc_z2x_add(b,b,c);
		CHECKTEST("division w/ quotient overlap",CORRECTNESS_PRC);
	}

	printf("%d tests %s ; %d bad inputs\n", CORRECTNESS_NTESTS,
			ok?"passed":"done ; failures encountered",
			fl);

	pop_time("Time for division tests");

	gmp_randclear(prng);

	mploc_z2_clear(scalar_elt);
	mploc_z2x_clear(x);
	mploc_z2x_clear(y);
	mploc_z2x_clear(z);
	mploc_z2x_clear(a);
	mploc_z2x_clear(b);
	mploc_z2x_clear(c);
}
#undef CHECKTEST

void	fixed_width_timings_z2x(mplf_t z2)
{
	int	i;
	mploc_z2x_t x,y,z,a,b,c;
	mpz_t	r;
	gmp_randstate_t prng;
	
	mploc_z2x_init(x,z2);
	mploc_z2x_init(y,z2);
	mploc_z2x_init(z,z2);
	mploc_z2x_init(a,z2);
	mploc_z2x_init(b,z2);
	mploc_z2x_init(c,z2);

	gmp_randinit_default(prng);
	
	mpz_init(r);

	timings_set_dispmode(TIMINGS_ARE_HUMAN_READABLE);

	printf("\nFixed-width timings (precision=%d)\n",FIXED_WIDTH_PRC);

	/**********************************************************/
	mploc_z2x_set_random(x,prng,CORRECTNESS_Z2X_DEGREE,FIXED_WIDTH_PRC);
	mploc_z2x_set_random(y,prng,CORRECTNESS_Z2X_DEGREE,FIXED_WIDTH_PRC);

	push_time();
	for(i=0;i<TIMINGS_NTESTS;i++) {
		mploc_z2x_add(z,x,y);
	}
	pop_time("fixed-width addition");

	/**********************************************************/
	mploc_z2x_set_random(x,prng,CORRECTNESS_Z2X_DEGREE,FIXED_WIDTH_PRC);
	mploc_z2x_set_random(y,prng,CORRECTNESS_Z2X_DEGREE,FIXED_WIDTH_PRC);

	push_time();
	for(i=0;i<TIMINGS_NTESTS;i++) {
		mploc_z2x_mul(z,x,y);
	}
	pop_time("fixed-width multiplication");

	gmp_randclear(prng);

	mploc_z2x_clear(x);
	mploc_z2x_clear(y);
	mploc_z2x_clear(z);
	mploc_z2x_clear(a);
	mploc_z2x_clear(b);
	mploc_z2x_clear(c);
}

void	variable_width_timings_z2x(mplf_t z2)
{
	int	i;
	mp_size_t	j;
	mploc_z2x_t x,y,z,a,b,c;
	mp_size_t curr_prec;
	gmp_randstate_t prng;
	
	mploc_z2x_init(x,z2);
	mploc_z2x_init(y,z2);
	mploc_z2x_init(z,z2);
	mploc_z2x_init(a,z2);
	mploc_z2x_init(b,z2);
	mploc_z2x_init(c,z2);

	gmp_randinit_default(prng);
	
	timings_set_dispmode(TIMINGS_ARE_SECONDS);

	printf("\nVariable-width timings\n");

	/**********************************************************/
	fprintf(stderr,"Benchmarking multiplication\n");


	for(j=1;j<1000;j++) {
		curr_prec = FIXED_WIDTH_PRC; //precisions_to_test[j];

		mploc_z2x_set_random(x,prng,j,curr_prec);
		mploc_z2x_set_random(y,prng,j,curr_prec);


		push_time();
		for(i=0;i<TIMINGS_NTESTS && look_time_as_sec()<0.1;i++) {
			mploc_z2x_mul(z,x,y);
			//mploc_z2x_normalize(z);
		}
		printf("%lu %u ",j,i);
		pop_time_raw();
		printf("\n");
	}

	gmp_randclear(prng);

	mploc_z2x_clear(x);
	mploc_z2x_clear(y);
	mploc_z2x_clear(z);
	mploc_z2x_clear(a);
	mploc_z2x_clear(b);
	mploc_z2x_clear(c);
}

int main(int argc, char * argv[])
{
	mplf_t	z2 UNUSED_VARIABLE;
	mplf_t	zp UNUSED_VARIABLE;
	mpz_t	p UNUSED_VARIABLE;
	int	i UNUSED_VARIABLE;

	setvbuf(stdout,NULL,_IONBF,0);
	setvbuf(stderr,NULL,_IONBF,0);

	mpz_init(seed);

	/*********************************/
	printf(MPLOC_BANNER "#### Zp test\n\n\n");

	/* Creation ; Selection of prime number. */
	if (argc != 3) {
		fprintf(stderr,"Need 2 args\n");
		exit(1);
	}

	mpz_init_set_ui(p,atoi(argv[1]));
	mpz_pow_ui(p,p,atoi(argv[2]));
	for(i=0;!mpz_probab_prime_p(p,5);i++) mpz_add_ui(p,p,1);
	printf("p=%d^%d + %d\n", atoi(argv[1]), atoi(argv[2]), i);

	zp=mplf_zp_create(p);

	correctness_zp(zp);

	mplf_zp_destroy(&zp);

	/*********************************/
	printf(MPLOC_BANNER "#### Z2 test\n\n\n");

	z2=mplf_z2_create();

	correctness_z2(z2);

	mplf_z2_destroy(&z2);

	/*********************************/
	printf(MPLOC_BANNER "#### Z2X test\n\n\n");

	z2=mplf_z2_create();

	correctness_z2x(z2);

	mplf_z2_destroy(&z2);

#if 0
	/*********************************/
	printf(MPLOC_BANNER "#### Zp bench\n\n\n");

	mpz_init_set_ui(p,atoi(argv[1]));
	mpz_pow_ui(p,p,atoi(argv[2]));
	for(i=0;!mpz_probab_prime_p(p,5);i++) mpz_add_ui(p,p,1);
	printf("p=%d^%d + %d\n", atoi(argv[1]), atoi(argv[2]), i);

	zp=mplf_zp_create(p);

	fixed_width_timings_zp(zp);
	variable_width_timings_zp(zp);

	mplf_zp_destroy(&zp);

	/*********************************/
	printf(MPLOC_BANNER "#### Z2 bench\n\n\n");

	z2=mplf_z2_create();

	fixed_width_timings_z2(z2);
	variable_width_timings_z2(z2);

	mplf_z2_destroy(&z2);
#endif

	/*********************************/
	printf(MPLOC_BANNER "#### Z2X bench\n\n\n");

	z2=mplf_z2_create();

	correctness_z2x(z2);
	fixed_width_timings_z2x(z2);
	variable_width_timings_z2x(z2);

	mplf_z2_destroy(&z2);

	mpz_clear(seed);

	return 0;
}

