/* $Id: timings.c 192 2004-10-13 14:15:09Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifdef  HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <math.h>
#include <assert.h>

#include "manu.h"
#include "timings.h"

#define MAX_TIMINGS_STACK	16

// typedef unsigned long long fasttimer_t;

static fasttimer_t timing_stack[MAX_TIMINGS_STACK];
static int stack_top;

static FILE * timings_file;

#define rdtsc(low,high) \
     __asm__ __volatile__("nop\nnop\nnop\nnop\nrdtsc" : "=a" (low), "=d" (high))

fasttimer_t nclocks()
{
        register unsigned int eax, edx;
        rdtsc(eax,edx);
        return ((fasttimer_t)edx)<<32|((fasttimer_t)eax);
}

/* Can be converted to something else */
fasttimer_t read_time(void)
{
	struct rusage res[1];
	getrusage(RUSAGE_SELF,res);
	return ((fasttimer_t)res->ru_utime.tv_sec)*1000000ULL+
		((fasttimer_t)res->ru_utime.tv_usec);
}

static double dispmode; /* =TIMINGS_ARE_HUMAN_READABLE = 0 */

void timings_set_dispmode(double mode)
{
	dispmode = mode;
}

int interpret_time(char *buf, size_t buflen, fasttimer_t t)
{
	double d;

	if (dispmode == TIMINGS_ARE_HUMAN_READABLE) {
		if (t<1000ULL) {
			return snprintf(buf,buflen,"%dus",(int)t);
		} else if (t<1000000ULL) {
			d=(double) t;
			d=d/1000.;
			return snprintf(buf,buflen,"%.2fms",d);
		} else if (t<60000000ULL) {
			t=t/1000ULL;
			d=(double) t;
			d=d/1000.;
			return snprintf(buf,buflen,"%.2fs",d);
		} else if (t<36000000000ULL) {
			int m,s;
			t=t/1000000ULL;
			s=(int)t % 60;
			m=(int)t / 60;
			return snprintf(buf,buflen,"%dmn%02ds",m,s);
		} else if (t<864000000000ULL) {
			int h,m;
			t=t/60000000ULL;
			m=(int)t % 60;
			h=(int)t / 60;
			return snprintf(buf,buflen,"%dhr%02dmn",h,m);
		} else {
			int d,h;
			t=t/3600000000ULL;
			h=(int)t % 24;
			d=(int)t / 24;
			return snprintf(buf,buflen,"%dd%02dhr",d,h);
		}
	} else {
		return snprintf(buf,buflen,"%.4e",((double)t)/dispmode);
	}
}

void set_timings_file(FILE * f)
{
	timings_file = f;
}

void push_time(void)
{
	if (stack_top==MAX_TIMINGS_STACK) BUG();

	timing_stack[stack_top++]=read_time();
}

static void avoid_zero_file(void)
{
	if (!timings_file) timings_file=stdout;
}

void pop_time_raw(void)
{
	char buf[20];

	if (stack_top==0) BUG();

	avoid_zero_file();
	
	interpret_time(buf,20,read_time()-timing_stack[--stack_top]);
	fprintf(timings_file,"%s",buf);
}

void discard_timer(void)
{
	--stack_top;
}

void pop_time_colon(const char * msg)
{
	avoid_zero_file();
	
	fprintf(timings_file,"%s: ", msg);
	pop_time_raw();
	fprintf(timings_file,"\n");
}

void estimate_arrival(FILE *f, char * msg, double fraction)
{
	char buf[20];

	if (stack_top==0) BUG();

	avoid_zero_file();

	interpret_time(buf,20,(fasttimer_t)ceil((read_time()-timing_stack[stack_top-1])/fraction));
	fprintf(f,"%s: %s\n",msg,buf);
}

double look_time_as_sec(void)
{
	if (stack_top==0) BUG();

	return (double)(((read_time()-timing_stack[stack_top-1])/1000ULL)/1000.);
}

fasttimer_t lookup_time(void)
{
	return read_time()-timing_stack[stack_top-1];
}

