/* $Id: agm.c 169 2004-09-30 14:39:30Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifdef  HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>

#include "mploc.h"
#include "z2.h"
#include "z2x.h"
#include "z2n.h"

/* Global variable for the basefield: beurk */
mplf_t myfield;


// a bad implementation of sqrt: works only if valuation is 0
#if 0
int mploc_z2n_sqrt(mploc_z2x_ptr z, mploc_z2x_srcptr x)
{
	int r;

	if (mploc_z2n_eq_ui(x, 0)) {
		mploc_z2n_set(z, x);
		return 1;
	}

	if (z != x) {
		r = mploc_z2n_invsqrt(z, x);
		if (r)
			mploc_z2n_mul(z, z, x);

		return r;
	} else {		/* need an aux variable if the pointers are the same */
		mploc_z2n_t aux;

		mploc_z2n_init(aux, myfield);

		r = mploc_z2n_invsqrt(aux, x);
		if (r)
			mploc_z2n_mul(z, aux, x);

		mploc_z2n_clear(aux);
		return r;
	}
}
#endif


void BivariateAGM(mploc_z2n_t res, mplf_t z2n, mploc_z2n_t a6)
{
	mploc_z2n_t a, b, a0, na, nb, aux;
	int k, maxk;
	int i, n;
	int prec;

	mploc_z2n_init(a, z2n);
	mploc_z2n_init(a0, z2n);
	mploc_z2n_init(b, z2n);
	mploc_z2n_init(na, z2n);
	mploc_z2n_init(nb, z2n);
	mploc_z2n_init(aux, z2n);

	maxk = (z2n->degree) / 2 + 4;
//  maxk = 2*maxk;
	prec = maxk + 1;

	/* init: a = 1 + 8*a6, b = 1 mod 16 */
	mploc_z2n_lshift(a, a6, 3);
	mploc_z2n_add_ui(a, a, 1);
	mploc_z2n_set_ui(b, 1, prec);

	/* AGM loop */
	printf("precision\n");
	for (k = 4; k <= maxk; ++k) {
		/* printf("precision %d\n", k); */
		mploc_z2n_force_precision(a, k + 2);
		mploc_z2n_force_precision(b, k + 2);
		mploc_z2n_add(na, a, b);
		mploc_z2n_rshift(na, na, 1);
		mploc_z2n_mul(nb, a, b);
		mploc_z2n_sqrt(nb, nb);

		mploc_z2n_sub_ui(aux, nb, 1);
		if (mploc_z2n_valuation(aux) < 2)
			mploc_z2n_neg(nb, nb);

		mploc_z2n_set(a, na);
		mploc_z2n_set(b, nb);
	}

	mploc_z2n_set(a0, a);

	/* one more complete loop */
	n = z2n->degree;

	printf("loop\n");
	for (i = 0; i < n; ++i) {
		/* printf("loop: %d\n", i); */
		mploc_z2n_force_precision(a, maxk + 2);
		mploc_z2n_force_precision(b, maxk + 2);
		mploc_z2n_add(na, a, b);
		mploc_z2n_rshift(na, na, 1);
		mploc_z2n_mul(nb, a, b);
		mploc_z2n_sqrt(nb, nb);

		mploc_z2n_sub_ui(aux, nb, 1);
		if (mploc_z2n_valuation(aux) < 2)
			mploc_z2n_neg(nb, nb);

		mploc_z2n_set(a, na);
		mploc_z2n_set(b, nb);
	}

	/* return the quotient a0/a */
	mploc_z2n_inv(a, a);
	mploc_z2n_mul(res, a, a0);


	mploc_z2n_clear(a);
	mploc_z2n_clear(a0);
	mploc_z2n_clear(b);
	mploc_z2n_clear(na);
	mploc_z2n_clear(nb);
	mploc_z2n_clear(aux);

}


void AGM(mplf_t z2n, mploc_z2n_t a6)
{
	mploc_z2n_t lambda, aux;
	int k, maxk;

	mploc_z2n_init(lambda, z2n);
	mploc_z2n_init(aux, z2n);

	maxk = (z2n->degree) / 2 + 4;

	/* lambda := 1 + 8*a6 */
	mploc_z2n_lshift(lambda, a6, 3);
	mploc_z2n_add_ui(lambda, lambda, 1);

	/* AGM loop:  lambda' := (1+lambda) / (2*sqrt(lambda)) */
	printf("precision\n");
	for (k = 4; k <= maxk; ++k) {
		/* printf("%d, lambda known at precision %lu\n", k,
		       lambda->prec); */
		mploc_z2n_force_precision(lambda, k + 2);
		mploc_z2n_add_ui(aux, lambda, 1);
		mploc_z2n_rshift(aux, aux, 1);
		mploc_z2n_invsqrt(lambda, lambda);
		mploc_z2n_mul(lambda, lambda, aux);
	}			/* end for */

	/* Compute Norm(2*lambda/(1+lambda)) */




	mploc_z2n_clear(lambda);
	mploc_z2n_clear(aux);

}

/* convert an element of Z_q into a signed integer in the Hasse-Weil
 * interval */
void printTraceInHW(mploc_z2n_t tr)
{
	mploc_t Tr;
	mplf_t z2;
	mpz_t zTr, modulus;
	int HWlim, sign;

	HWlim = (myfield->degree) / 2 + 4;

	z2 = mplf_z2_create();
	mploc_z2_init(Tr, z2);

	mploc_z2x_get_coeff(Tr, tr, 0);
	mpz_init(zTr);
	mpz_set(zTr, Tr->x);
	mpz_init(modulus);
	mpz_set_ui(modulus, 1);
	mpz_mul_2exp(modulus, modulus, HWlim);
	mpz_mod(zTr, zTr, modulus);
	mpz_set_ui(modulus, 1);
	mpz_mul_2exp(modulus, modulus, HWlim - 1);
	if (mpz_cmp(zTr, modulus) > 0) {
		sign = -1;
		mpz_mul_ui(modulus, modulus, 2);
		mpz_sub(zTr, modulus, zTr);
	} else {
		sign = 1;
	}

	if (sign == -1)
		printf("-");
	mpz_out_str(stdout, 10, zTr);

	mpz_clear(modulus);
	mpz_clear(zTr);
	mploc_z2_clear(Tr);
	mplf_z2_destroy(&z2);
}


int main()
{
	mploc_z2x_t poldef;
	mploc_t z2_one;
	mplf_t myz2;
	mp_size_t prec;
	gmp_randstate_t prng;

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	gmp_randinit_default(prng);

	/* create Z_2 */
	myz2 = mplf_z2_create();
	/* create and initialize poldef */
	mploc_z2x_init(poldef, myz2);	/* necessary to build myz2 first??? */
	prec = 100;
#if 1
	/* poldef = x^163 + x^7 + x^6 + x^3 + 1 */
	mploc_z2x_set_xi(poldef, 163, prec);
	mploc_z2x_add_ui(poldef, poldef, 1);
	mploc_z2_init(z2_one, myz2);
	mploc_z2_set_ui(z2_one, 1, prec);
	mploc_z2x_set_coeff(poldef, 3, z2_one);
	mploc_z2x_set_coeff(poldef, 6, z2_one);
	mploc_z2x_set_coeff(poldef, 7, z2_one);
#else
	/* poldef = x^11 + x^2 + 1 */
	mploc_z2x_set_xi(poldef, 11, prec);
	mploc_z2x_add_ui(poldef, poldef, 1);
	mploc_z2_init(z2_one, myz2);
	mploc_z2_set_ui(z2_one, 1, prec);
	mploc_z2x_set_coeff(poldef, 2, z2_one);
#endif

	printf("defpol=");
	mploc_z2x_print(poldef);
	printf("\n");

	/* create z_2^n */
	myfield = mplf_z2n_create(poldef);

	/* REAL STUFF IS HERE !!!! */
	{
		mploc_z2n_t a6, tr;
		mploc_z2n_init(a6, myfield);
		mploc_z2n_init(tr, myfield);

		mploc_z2n_set_random(a6, prng, 1);
		printf("a6=");
		mploc_z2n_print(a6);
		printf("\n");

		AGM(myfield, a6);

		BivariateAGM(tr, myfield, a6);
		/*
		   mploc_z2n_print(tr);
		   printf("\n");
		 */
		printTraceInHW(tr);
		printf("\n");

		mploc_z2n_clear(a6);
		mploc_z2n_clear(tr);
	}

	/* With default random source, I get:
defpol=1 + X^3 + X^6 + X^7 + X^163 + O(2^100)
a6=1 + X^3 + X^5 + X^8 + X^9 + X^10 + X^12 + X^13 + X^15 + X^16 + X^17 + X^18 + X^19 + X^20 + X^21 + X^22 + X^25 + X^29 + X^33 + X^36 + X^38 + X^39 + X^40 + X^41 + X^44 + X^47 + X^48 + X^49 + X^51 + X^55 + X^57 + X^60 + X^61 + X^62 + X^63 + X^64 + X^67 + X^70 + X^71 + X^72 + X^73 + X^74 + X^77 + X^78 + X^79 + X^84 + X^87 + X^88 + X^89 + X^90 + X^91 + X^94 + X^100 + X^102 + X^103 + X^104 + X^106 + X^107 + X^108 + X^111 + X^115 + X^117 + X^124 + X^128 + X^129 + X^130 + X^131 + X^132 + X^134 + X^136 + X^141 + X^142 + X^145 + X^146 + X^150 + X^154 + X^155 + X^158 + X^161 + O(2^1)
161088514449475591064501
	*/

	REALLOC_INFO();

	/* free everything */
	mplf_z2n_destroy(&myfield);
	mploc_z2x_clear(poldef);
	mploc_z2_clear(z2_one);
	mplf_z2_destroy(&myz2);
	return 0;
}
