/* $Id: teich.c 205 2006-09-29 14:42:37Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>
#ifdef  HAVE_GMP_GMP_MPARAM_H
#include <gmp/gmp-mparam.h>
#elif defined(HAVE_GMP_MPARAM_H)
#include <gmp-mparam.h>
#endif

#include <assert.h>
#include <string.h>

#include "manu.h"

#include "mploc.h"
#include "z2.h"
#include "z2x.h"
#include "z2n.h"

#include "teich-common.h"
#include "irr_tbl.h"
#include "teich1.h"
#include "teich2.h"

#define	DEFAULT_W_FOR_SST	GMP_LIMB_BITS
#define	DEFAULT_DEG		7
#define	DEFAULT_TARGET		128

mplf_t z2;

void usage(int rc)
{
	fprintf(stderr, "Usage: teich [--rec] <deg> <target> [<w>]\n");
	exit(rc);
}

int main(int argc, char * argv[])
{
	mploc_z2x_t f;
	int deg=DEFAULT_DEG;
	mp_size_t target=DEFAULT_TARGET;
	mp_size_t w=(mp_limb_t) -1;
	int p;
	int rec=0;

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	z2 = mplf_z2_create();
	mploc_z2x_init(f, z2);

	for(p=0;argc>1;argv++,argc--) {
		if (strcmp(argv[1],"--help")==0) {
			usage(0);
		}
		if (strcmp(argv[1],"--rec")==0) {
			rec=1;
			continue;
		}
		mp_size_t v=atoi(argv[1]);
		if (v==0) usage(1);
		switch (p++) {
			case 0: deg=v; break;
			case 1: target=v; break;
			case 2: w=v; break;
			default: usage(1);
		}
	}

	if (w==(mp_limb_t) -1) {
		w=DEFAULT_W_FOR_SST;
	} else {
		if (rec) usage(1);
	}

	if (!mploc_z2x_set_irrpol_db_quick(f,deg)) {
		fprintf(stderr,"Degree %d poly is not in irr_tbl\n", deg);
		exit(1);
	}
		
	if (rec) {
		mploc_z2x_t tmp;
		fprintf(stderr,"Doing recursive computation\n");
		mploc_z2x_init(tmp,z2);
		mploc_z2x_canonical_lift(tmp,f,target);
		mploc_z2x_set_change_precision(f,tmp,target);
		mploc_z2x_clear(tmp);
	} else {
		fprintf(stderr,"Doing SST computation\n");
		mploc_z2x_canonical_lift_sst(f,f,target,w);
		mploc_z2x_force_precision(f,target);
	}

	/*
	printf("R<X>:=PolynomialRing(pAdicQuotientRing(2,%ld));\n",f->prec);
	printf("O:=func<x|R!0>;\n");
	mploc_z2x_print(f);
	printf(" eq SSTLiftCyclotomicPolynomialNew(%d,%ld,1);\n",f->deg,f->prec);
	*/
	mploc_z2x_print(f);
	printf("\n");

	REALLOC_INFO();

	mploc_z2x_clear(f);
	mplf_z2_destroy(&z2);

	return 0;
}
