/* $Id: teich1.c 169 2004-09-30 14:39:30Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifdef  HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <gmp.h>

#include <assert.h>

#include "manu.h"

#include "mploc.h"
#include "z2.h"
#include "z2x.h"
#include "z2n.h"

#include "teich-common.h"
#include "teich1.h"

/* evaluates f(-x) */
void twist_poly(mploc_z2x_ptr g, mploc_z2x_srcptr f)
{
	int i;
	mploc_t c;
	mploc_z2_init(c,z2);

	mploc_z2x_set_xi(g,f->deg,f->prec);
	for(i=0;i<=f->deg;i++) {
		mploc_z2x_get_coeff(c,f,i);
		if (i&1) mploc_z2_neg(c,c);
		mploc_z2x_set_coeff(g,i,c);
	}
	
	mploc_z2_clear(c);
}

/* This gives the error term (well, the opposite in fact). The first bit
 * is correct, the others must be adjsuted by SST.
 */
void teich_get_phi(mploc_z2x_ptr v, mploc_z2x_srcptr f)
{
	mploc_z2x_t f_sqr,f_neg;
	mploc_t c;
	int i;
	mploc_z2x_init(f_sqr,z2);
	mploc_z2x_init(f_neg,z2);
	mploc_z2_init(c,z2);
	/* This is just a way to to realloc & zero simultaneously */
	mploc_z2x_set_xi(f_sqr,2*f->deg,f->prec);

	twist_poly(f_neg,f);

	for(i=0;i<=f->deg;i++) {
		mploc_z2x_get_coeff(c,f,i);
		mploc_z2x_set_coeff(f_sqr,2*i,c);
	}

	mploc_z2x_mul(v,f,f_neg);
	if ((f->deg & 1) == 0) {
		mploc_z2x_sub(v,f_sqr,v);
	} else {
		mploc_z2x_add(v,f_sqr,v);
	}

	mploc_z2x_clear(f_sqr);
	mploc_z2x_clear(f_neg);

	if (v->deg==-1) {
		mploc_z2_clear(c);
		return;
	}
	/* v may be zero ! This can happen if the current approximation
	 * is more correct than expected (see approximations of X^7+X+1
	 * at precisions 11 and 12). */
	ASSERT((v->deg&1)==0);
	for(i=0;i<=v->deg;i++) {
		mploc_z2x_get_coeff(c,v,i);
		if (i&1) {
			ASSERT(mploc_z2_eq_ui(c,0));
			continue;
		}
		mploc_z2x_set_coeff(v,i>>1,c);
	}
	v->deg>>=1;
	ASSERT(mploc_z2x_is_good_looking(v));
	mploc_z2_clear(c);
}

/* The L operator */
void teich_sst_helper(
		mploc_z2x_ptr y,
		mploc_z2x_srcptr f_neg,
		mploc_z2x_srcptr e,
		mp_size_t w)
{
	mploc_t c;
	int i;

	mploc_z2_init(c,z2);
	ASSERT(e->prec==w);
	ASSERT(f_neg->prec==w);
	mploc_z2x_set_change_precision(y,e,w);
	mploc_z2x_mul(y,f_neg,y);
	for(i=0;i<=y->deg;i++) {
		mploc_z2x_get_coeff(c,y,i);
		if (i&1) continue;
		mploc_z2x_set_coeff(y,i>>1,c);
	}
	y->deg>>=1;
	/* In some cases, if y->deg is odd, we can experience cancelation
	 * of the middle terms which become leading later on.
	 *
	 * (such problems would not happen if we were adding the product
	 * f * e(-x) as well).
	 */
	mploc_z2x_normalize(y);
	mploc_z2_clear(c);
}

void mploc_z2x_canonical_lift_sst(mploc_z2x_ptr f, mploc_z2x_srcptr src,
		mp_size_t target, mp_size_t w)

{
	mploc_z2x_t v;
	mploc_z2x_t g,f_neg;	/* used by sst code */
	int i,k;

	mploc_z2x_init(v,z2);
	mploc_z2x_init(g,z2);
	mploc_z2x_init(f_neg,z2);

	mploc_z2x_set(f,src);

	/* start this the dumb way */

	for(i=1;i<w;i++) {
		mp_limb_t oprec;
		
		oprec=f->prec;
		mploc_z2x_force_precision(f,oprec+1);
		teich_get_phi(v,f);
		ASSERT(mploc_z2x_valuation(v) >= oprec);
		mploc_z2x_sub(f,f,v);
	}

	/*
	printf("R<X>:=PolynomialRing(pAdicQuotientRing(2,%ld));\n",f->prec);
	printf("O:=func<x|R!0>;\n");
	mploc_z2x_print(f);
	printf(" eq SSTLiftCyclotomicPolynomialNew(%d,%ld,1);\n",f->deg,f->prec);
	*/

	twist_poly(f_neg,f);
	if (f->deg&1)
		mploc_z2x_neg(f_neg,f_neg);

	for(k=1;k<iceildiv(target,w);k++) {
		mp_limb_t oprec;
		int j;

		oprec=f->prec;
		ASSERT(oprec == k*w);
		mploc_z2x_force_precision(f,f->prec+w);
		teich_get_phi(v,f);
		ASSERT(mploc_z2x_valuation(v) >= oprec);

		mploc_z2x_rshift(v,v,k*w);
		ASSERT(v->prec==w);

		mploc_z2x_set(g,v);
		for(j=1;j<w;j++) {
			teich_sst_helper(v,f_neg,v,w);
			mploc_z2x_lshift_quick(v,v,1);
			mploc_z2x_add(g,g,v);
		}

		mploc_z2x_lshift(g,g,k*w);
		mploc_z2x_sub(f,f,g);
	}

	mploc_z2x_clear(v);
	mploc_z2x_clear(g);
	mploc_z2x_clear(f_neg);
}
