#ifndef X86INTER_H_
#define X86INTER_H_

/* The asm inlines from this files are generic-looking, but the command
 * set around which they're built is really too x86-ish, so they end up
 * in the x86 directory
 */

#define MUL_4_BY_0      "0"
#define MUL_4_BY_1      "4"
#define MUL_4_BY_2      "8"
#define MUL_4_BY_3      "12"
#define MUL_4_BY_4      "16"
#define MUL_4_BY_5      "20"
#define MUL_4_BY_6      "24"
#define MUL_4_BY_7      "28"

#define A(i)    MUL_4_BY_ ## i "(%[s1])"
#define B(i)    MUL_4_BY_ ## i "(%[s2])"
#define R(i)    MUL_4_BY_ ## i "(%%ebp)"
#define	S(i)	MUL_4_BY_ ## i "(%[s])"
#define	T(i)	MUL_4_BY_ ## i "(%[t])"
#define rB      "%%ebx"
#define rC      "%%ecx"
#define CN(x,n,c)       [x] # c (x)

/* The header below includes the length command set (or, bunch of asm
 * shortcuts, more properly).
 */
#include "asm/x86/x86inter_cset.h"

/* Does 3 (three) 32x32 -> 64 mults ; No carry handling is needed */
ASM_FUNCTION(MUL2,__x86_inter_1)
INLINER	void MUL2__x86_inter_1(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	asm volatile(
		SAVE_EBP_AND_SET
		m_put_put(A(0),B(0),R(0),rB)

		mc_put(A(1),B(0),rB,rC)
		mc_put(A(0),B(1),rC,R(1))

		RESTORE_EBP
		: : CN(r,0,m), CN(s1,1,S),  CN(s2,2,D):
		"eax","edx","ebx","ecx");
}

ASM_FUNCTION(MUL3,__x86_inter_1)
INLINER void MUL3__x86_inter_1(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	asm volatile(
		SAVE_EBP_AND_SET

		m_put_put(A(0),B(0),R(0),rB)

		mc_put_put(A(1),B(0),rB,R(1),rB)
		mc_put(A(2),B(0),rB,R(2))

		m_put_put(A(0),B(1),rC,rB)
		x__I_fm(A(1),B(1),R(1),R(2))
		
		m_add(A(0),B(2),R(2))

		RESTORE_EBP
		: : CN(r,0,m), CN(s1,1,S),  CN(s2,2,D):
		"eax","edx","ebx","ecx");
}

/* Fifteen mults here. 11 could be achieved by 4+1 karatsuba. */
/* Since carry handling becomes cumbersome here, we're now
 * destination-local instead of source-local. It isn't much different in
 * terms of performance anyway.
 */
ASM_FUNCTION(MUL4,__x86_inter_1)
INLINER void MUL4__x86_inter_1(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	MUL_NO_OVERLAP();

	asm volatile(
			/* This is source-local */
		SAVE_EBP_AND_SET

		m_put_put(A(0),B(0),R(0),rB)

		mc_put_put(A(1),B(0),rB,R(1),rB)
		mc_put_put(A(2),B(0),rB,R(2),rB)
		mc_put(A(3),B(0),rB,R(3))

		m_put_put(A(0),B(1),rC,rB)
		x__I_sm(A(1),B(1),R(1))
		x_I__fm(A(2),B(1),R(2),R(3))
		
		m_put_put(A(0),B(2),rC,rB)
		x__I_fm(A(1),B(2),R(2),R(3))
		
		m_add(A(0),B(3),R(3))

		RESTORE_EBP
		: : CN(r,0,m), CN(s1,1,S),  CN(s2,2,D):
		"eax","edx","ebx","ecx");
}

/* This asm version gives 87 cycles or so... */
ASM_FUNCTION(MUL5,__x86_inter_2)
INLINER void MUL5__x86_inter_2(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	MUL_NO_OVERLAP();
	asm volatile(
			/* This is source-local */
		SAVE_EBP_AND_SET

		m_put_put(A(0),B(0),R(0),rB)

		mc_put_put(A(1),B(0),rB,R(1),rB)
		mc_put_put(A(2),B(0),rB,R(2),rB)
		mc_put_put(A(3),B(0),rB,R(3),rB)
		mc_put(A(4),B(0),rB,R(4))

		m_put_put(A(0),B(1),rC,rB)
		x__I_sm(A(1),B(1),R(1))
		x_I__sm(A(2),B(1),R(2))
		x__I_fm(A(3),B(1),R(3),R(4))
		
		m_put_put(A(0),B(2),rC,rB)
		x__I_sm(A(1),B(2),R(2))
		x_I__fm(A(2),B(2),R(3),R(4))
		
		m_put_put(A(0),B(3),rC,rB)
		x__I_fm(A(1),B(3),R(3),R(4))
		
		m_add(A(0),B(4),R(4))

		RESTORE_EBP
		: : CN(r,0,m), CN(s1,1,S),  CN(s2,2,D):
		"eax","edx","ebx","ecx");
}

ASM_FUNCTION(MUL6,__x86_inter_0)
INLINER void MUL6__x86_inter_0(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	MUL_NO_OVERLAP();

	asm volatile(
			/* This is source-local */
		SAVE_EBP_AND_SET

		m_put_put(A(0),B(0),R(0),rB)

		mc_put_put(A(1),B(0),rB,R(1),rB)
		mc_put_put(A(2),B(0),rB,R(2),rB)
		mc_put_put(A(3),B(0),rB,R(3),rB)
		mc_put_put(A(4),B(0),rB,R(4),rB)
		mc_put(A(5),B(0),rB,R(5))

		m_put_put(A(0),B(1),rC,rB)
		x__I_sm(A(1),B(1),R(1))
		x_I__sm(A(2),B(1),R(2))
		x__I_sm(A(3),B(1),R(3))
		x_I__fm(A(4),B(1),R(4),R(5))
		
		m_put_put(A(0),B(2),rC,rB)
		x__I_sm(A(1),B(2),R(2))
		x_I__sm(A(2),B(2),R(3))
		x__I_fm(A(3),B(2),R(4),R(5))
		
		m_put_put(A(0),B(3),rC,rB)
		x__I_sm(A(1),B(3),R(3))
		x_I__fm(A(2),B(3),R(4),R(5))
		
		m_put_put(A(0),B(4),rC,rB)
		x__I_fm(A(1),B(4),R(4),R(5))
		
		m_add(A(0),B(5),R(5))

		RESTORE_EBP
		: : CN(r,0,m), CN(s1,1,S),  CN(s2,2,D):
		"eax","edx","ebx","ecx");
}

ASM_FUNCTION(MUL7,__x86_inter_0)
INLINER void MUL7__x86_inter_0(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	MUL_NO_OVERLAP();

	asm volatile(
			/* This is source-local */
		SAVE_EBP_AND_SET

		m_put_put(A(0),B(0),R(0),rB)

		mc_put_put(A(1),B(0),rB,R(1),rB)
		mc_put_put(A(2),B(0),rB,R(2),rB)
		mc_put_put(A(3),B(0),rB,R(3),rB)
		mc_put_put(A(4),B(0),rB,R(4),rB)
		mc_put_put(A(5),B(0),rB,R(5),rB)
		mc_put(A(6),B(0),rB,R(6))

		m_put_put(A(0),B(1),rC,rB)
		x__I_sm(A(1),B(1),R(1))
		x_I__sm(A(2),B(1),R(2))
		x__I_sm(A(3),B(1),R(3))
		x_I__sm(A(4),B(1),R(4))
		x__I_fm(A(5),B(1),R(5),R(6))
		
		m_put_put(A(0),B(2),rC,rB)
		x__I_sm(A(1),B(2),R(2))
		x_I__sm(A(2),B(2),R(3))
		x__I_sm(A(3),B(2),R(4))
		x_I__fm(A(4),B(2),R(5),R(6))
		
		m_put_put(A(0),B(3),rC,rB)
		x__I_sm(A(1),B(3),R(3))
		x_I__sm(A(2),B(3),R(4))
		x__I_fm(A(3),B(3),R(5),R(6))
		
		m_put_put(A(0),B(4),rC,rB)
		x__I_sm(A(1),B(4),R(4))
		x_I__fm(A(2),B(4),R(5),R(6))
		
		m_put_put(A(0),B(5),rC,rB)
		x__I_fm(A(1),B(5),R(5),R(6))
		
		m_add(A(0),B(6),R(6))

		RESTORE_EBP
		: : CN(r,0,m), CN(s1,1,S),  CN(s2,2,D):
		"eax","edx","ebx","ecx");
}

ASM_FUNCTION(MUL8,__x86_inter_0)
INLINER void MUL8__x86_inter_0(mp_limb_t * r,const mp_limb_t * s1,const mp_limb_t * s2)
{
	MUL_NO_OVERLAP();

	asm volatile(
			/* This is source-local */
		SAVE_EBP_AND_SET

		m_put_put(A(0),B(0),R(0),rB)

		mc_put_put(A(1),B(0),rB,R(1),rB)
		mc_put_put(A(2),B(0),rB,R(2),rB)
		mc_put_put(A(3),B(0),rB,R(3),rB)
		mc_put_put(A(4),B(0),rB,R(4),rB)
		mc_put_put(A(5),B(0),rB,R(5),rB)
		mc_put_put(A(6),B(0),rB,R(6),rB)
		mc_put(A(7),B(0),rB,R(7))

		m_put_put(A(0),B(1),rC,rB)
		x__I_sm(A(1),B(1),R(1))
		x_I__sm(A(2),B(1),R(2))
		x__I_sm(A(3),B(1),R(3))
		x_I__sm(A(4),B(1),R(4))
		x__I_sm(A(5),B(1),R(5))
		x_I__fm(A(6),B(1),R(6),R(7))
		
		m_put_put(A(0),B(2),rC,rB)
		x__I_sm(A(1),B(2),R(2))
		x_I__sm(A(2),B(2),R(3))
		x__I_sm(A(3),B(2),R(4))
		x_I__sm(A(4),B(2),R(5))
		x__I_fm(A(5),B(2),R(6),R(7))
		
		m_put_put(A(0),B(3),rC,rB)
		x__I_sm(A(1),B(3),R(3))
		x_I__sm(A(2),B(3),R(4))
		x__I_sm(A(3),B(3),R(5))
		x_I__fm(A(4),B(3),R(6),R(7))
		
		m_put_put(A(0),B(4),rC,rB)
		x__I_sm(A(1),B(4),R(4))
		x_I__sm(A(2),B(4),R(5))
		x__I_fm(A(3),B(4),R(6),R(7))
		
		m_put_put(A(0),B(5),rC,rB)
		x__I_sm(A(1),B(5),R(5))
		x_I__fm(A(2),B(5),R(6),R(7))
		
		m_put_put(A(0),B(6),rC,rB)
		x__I_fm(A(1),B(6),R(6),R(7))
		
		m_add(A(0),B(7),R(7))

		RESTORE_EBP
		: : CN(r,0,m), CN(s1,1,S),  CN(s2,2,D):
		"eax","edx","ebx","ecx");
}

#undef	MUL_4_BY_0
#undef	MUL_4_BY_1
#undef	MUL_4_BY_2
#undef	MUL_4_BY_3
#undef	MUL_4_BY_4
#undef	MUL_4_BY_5
#undef	MUL_4_BY_6
#undef	MUL_4_BY_7
#undef	A
#undef	B
#undef	R
#undef	S
#undef	T
#undef	CN
#undef	rB
#undef	rC

/* These were defined by x86inter_cset.h */

#undef	m_put_put
#undef	m_add
#undef	mc_put_put
#undef	mc_put
#undef	x_I__sm
#undef	x__I_sm
#undef	x_I__fm
#undef	x__I_fm
#undef	MUL_RA_AC_P
#undef	MUL_RA_AC_MV
#undef	RESTORE_EBP
#undef	SAVE_EBP_AND_SET

#endif	/* X86INTER_H_ */
