#ifndef X86KERNEL_H_
#define X86KERNEL_H_

#ifdef __cplusplus
extern "C" {
#endif

/* Parts of the code here were taken (20021226) from gmp's longlong.h ;
 * there's a lot more there of course.
 */

#define	KERNEL_HAS_UMUL_PPMM
#define umul_ppmm(w1, w0, u, v)		\
  __asm__ ("mull	%3"		\
	   : "=a" (w0), "=d" (w1)	\
	   : "%0" ((mp_limb_t)(u)),	\
	     "rm" ((mp_limb_t)(v)))
	  /* 20021226: changing "rm" into "r" for the last constraint
	   * doesn't help much. */

#if 0
	  /* w0 and w1 are earlyclobbers here, but only wrt the carry
	   * argument. Therefore marking them as earlyclobbers is
	   * counter-productive, since it prevents v from going into an
	   * eax or edx register.
	   *
	   * The version below is preferred, perhaps gcc does better that
	   * way.
	   */
#define	umul_ppmm_and_carry(w1,w0,u,v,c)	\
  __asm__ (	"mull	%3\n\t"			\
		"addl	%4, %%eax\n\t"		\
		"adcl	$0, %%edx\n"		\
	   : "=&a" (w0), "=&d" (w1)					\
	   : "%0" ((mp_limb_t)(u)), "rm" ((mp_limb_t)(v)), "rm" ((mp_limb_t)c))
#else
#define	KERNEL_HAS_UMUL_PPMM_C
#define	umul_ppmm_and_carry(w1,w0,u,v,c)	\
  	BEGIN_BLOCK				\
  	umul_ppmm(w1,w0,u,v);			\
	carry_terminate_2(w1,w0,c);		\
	END_BLOCK
#endif

#if 0
#define umul_ppmm_c_noprop(w1, w0, u, v, c) \
  asm ("mull	%3\n	addl	%4, %0\n	adcl	$0,%1\n"	\
		  : "=&a" (w0), "=&d" (w1) : "%0" (u), "rm" (v), "rm" (c))

#define umul_ppmm_c_prop(w1, w0, u, v, c) \
  asm ("mull	%3\n	addl	%4, %0\n	adcl	$0,%1\n"	\
		  : "=&a" (w0), "=&d" (w1) : "%0" (u), "rm" (v), "rm" (c))

#define carry_propagate_1(w0,c)	\
  asm (	"	addl	%0, %1\n"		\
	"	adcl	$0, %0\n"		\
		  : "+rm" (c), "+rm" (w0))

#define carry_propagate_2(w1,w0,c)	\
  asm (	"	addl	%0, %1\n"		\
	"	adcl	$0, %2\n"		\
	"	adcl	$0, %0\n"		\
		  : "+rm" (c), "+rm" (w0), "+rm" (w1))

#define carry_propagate_3(w2,w1,w0,c)	\
  asm (	"	addl	%0, %1\n"		\
	"	adcl	$0, %2\n"		\
	"	adcl	$0, %3\n"		\
	"	adcl	$0, %0\n"		\
		  : "+rm" (c), "+rm" (w0), "+rm" (w1), "+rm" (w2))
#endif

#if 0
#define carry_terminate_1(w0,c)	\
  asm (	"	addl	%1, %0\n"		\
		  : "+rm,r!" (w0) : "r,m" (c))
#endif

#define	KERNEL_HAS_CTERM_2
#define carry_terminate_2(w1,w0,c)	\
  asm ( "	addl	%2, %0\n"		\
	"	adcl	$0, %1\n"		\
		  : "+r" (w0), "+rm" (w1) : "rm" (c))

  // for some unknown reason, adding any of these alternative constraints
  // breaks the results. (not only the carry. The whole thing breaks
  // down).  That's especially true for the second occurrence within MUL4
  
		  //: "+r,m!" (w0), "+rm,rm" (w1) : "rm,r" (c))
		  //: "+r,rm" (w0), "+rm,rm" (w1) : "rm,r" (c))

#define	KERNEL_HAS_CTERM_3
#define carry_terminate_3(w2,w1,w0,c)	\
  asm (	"	addl	%3, %0\n"		\
	"	adcl	$0, %1\n"		\
	"	adcl	$0, %2\n"		\
		  : "+r" (w0), "+rm" (w1), "+rm" (w2) : "rm" (c))

#define	KERNEL_HAS_CTERM_4
#define carry_terminate_4(w3,w2,w1,w0,c)	\
  asm (	"	addl	%4, %1\n"		\
	"	adcl	$0, %2\n"		\
	"	adcl	$0, %3\n"		\
	"	adcl	$0, %4\n"		\
		  : "+r"(w3), "+r" (w0), "+rm" (w1), "+rm" (w2) : "rm" (c))

#if 0
  // Disable insns with multiple alternative constraints.
#define add2_terminate(w1,w0,v1,v0)	\
  asm ( "	addl	%3, %1\n"	\
	"	adcl	%2, %0\n"	\
		: "+r,m??" (w1), "+r,m" (w0) : "rm,r"(v1), "rm,r"(v0))

#define add2_propagate(c,w1,w0,v1,v0)	\
  asm ( "	addl	%4, %2\n"	\
	"	adcl	%3, %1\n"	\
	"	adcl	$0, %0\n"	\
		: "=r,m?" (c), "+r,rm" (w1), "+r,rm" (w0): "rm,r"(v1), "rm,r"(v0))
#endif

#define	KERNEL_HAS_DBL2_TERM
#define double2_terminate(w1,w0)	\
  asm ( "	addl	%1, %1\n"	\
	"	adcl	%0, %0\n"	\
		: "+r" (w1), "+r" (w0))

#define	KERNEL_HAS_DBL3_TERM
#define double3_terminate(w2,w1,w0)	\
  asm ( "	addl	%2, %2\n"	\
	"	adcl	%1, %1\n"	\
	"	adcl	%0, %0\n"	\
		: "+r" (w2), "+r" (w1), "+r" (w0))

#define	KERNEL_HAS_DBL3_PROP_ADD
#define double3_propagate_add(c,w2,w1,w0)	\
  asm ( "	addl	%3, %3\n"	\
	"	adcl	%2, %2\n"	\
	"	adcl	%1, %1\n"	\
	"	adcl	$0, %0\n"	\
		: "+rm"(c), "+r" (w2), "+r" (w1), "+r" (w0))

#if 0
#define double2_propagate(c,w1,w0)	\
  asm ( "	addl	%2, %2\n"	\
	"	adcl	%1, %1\n"	\
	"	adcl	$0, %0\n"	\
		: "=r" (c), "+r" (w1), "+r" (w0))
		//: "=r,m?" (c), "+r,r" (w1), "+r,r" (w0))
#endif

/* In order to code the multiplication in C, we need a
 * macro doing the following: 
 def    macro(d1,d0,s1,s0,in)	\
	d0+=in;				\
	d1+=s0+CARRY_OUT;		\
	s1+=CARRY_OUT;

	The next ``stream data'' is s1(high),d1(low)

 * It is assumed that s2,s1 is <= w^2-2w+1, hence adding w-1 + carryout
 * incurs no external carry.
 */

#define	KERNEL_HAS_ADD2_STREAM
#define add2_stream(d1,d0,s1,s0,in)	\
  asm (	"	addl	%4,%1\n"	\
	"	adcl	%3,%0\n"	\
	"	adcl	$0,%2\n"	\
		: "+r" (d1), "+r" (d0), "+r" (s1) : "rm" (s0), "rm" (in))


/***************************************************************************/

/* count_leading_zeros stuff */
/* All of this is taken from gmp's longlong.h, with comments stripped out */

/* Define stuff for longlong.h.  */
typedef		 long	SItype;
typedef unsigned long	USItype;
typedef mp_limb_t	UWtype;
typedef unsigned int	UHWtype;
#define W_TYPE_SIZE	GMP_LIMB_BITS

#define __clz_tab               __MPN(clz_tab)

#if HAVE_HOST_CPU_i586 || HAVE_HOST_CPU_pentium
#define count_leading_zeros(c,n)					\
  do {									\
    union {								\
      double    d;							\
      unsigned  a[2];							\
    } __u;								\
    ASSERT ((n) != 0);							\
    __u.d = (UWtype) (n);						\
    (c) = 0x3FF + 31 - (__u.a[1] >> 20);				\
  } while (0)
#define COUNT_LEADING_ZEROS_0   (0x3FF + 31)

#else /* ! pentium */

#if HAVE_HOST_CPU_pentiummmx

#define count_leading_zeros(c,n)					\
  do {									\
    USItype  __n = (n);							\
    USItype  __shift;							\
    __asm__ ("cmpl  $0x1000000, %1\n"					\
	     "sbbl  %0, %0\n"						\
	     "cmpl  $0x10000, %1\n"					\
	     "sbbl  $0, %0\n"						\
	     "cmpl  $0x100, %1\n"					\
	     "sbbl  $0, %0\n"						\
	     : "=&r" (__shift) : "r"  (__n));				\
    __shift = __shift*8 + 24 + 1;					\
    (c) = 32 + 1 - __shift - __clz_tab[__n >> __shift];			\
  } while (0)

#define COUNT_LEADING_ZEROS_NEED_CLZ_TAB
#define COUNT_LEADING_ZEROS_0   31

#else /* !pentiummmx */

#if __GNUC__ < 3							\
  && (HAVE_HOST_CPU_i386                                                \
      || HAVE_HOST_CPU_i686                                             \
      || HAVE_HOST_CPU_pentiumpro                                       \
      || HAVE_HOST_CPU_pentium2                                         \
      || HAVE_HOST_CPU_pentium3)
#define count_leading_zeros(count, x)					\
  do {									\
    USItype __cbtmp;							\
    ASSERT ((x) != 0);							\
    __asm__ ("bsrl %1,%0" : "=r" (__cbtmp) : "rm" ((USItype)(x)));	\
    (count) = 31 - __cbtmp;						\
  } while (0)
#else
#define count_leading_zeros(count, x)					\
  do {									\
    USItype __cbtmp;							\
    ASSERT ((x) != 0);							\
    __asm__ ("bsrl %1,%0" : "=r" (__cbtmp) : "rm" ((USItype)(x)));	\
    (count) = __cbtmp ^ 31;						\
  } while (0)
#endif

#define count_trailing_zeros(count, x)					\
  do {									\
    ASSERT ((x) != 0);							\
    __asm__ ("bsfl %1,%0" : "=r" (count) : "rm" ((USItype)(x)));	\
  } while (0)
#endif /* ! pentiummmx */
#endif /* ! pentium */

#if !defined (count_trailing_zeros)
#define count_trailing_zeros(count, x) \
  do {									\
    UWtype __ctz_x = (x);						\
    UWtype __ctz_c;							\
    ASSERT (__ctz_x != 0);						\
    count_leading_zeros (__ctz_c, __ctz_x & -__ctz_x);			\
    (count) = W_TYPE_SIZE - 1 - __ctz_c;				\
  } while (0)
#endif


#ifdef __cplusplus
}
#endif

#endif	/* X86KERNEL_H_ */
