/* $Id: z2_cxx.hpp 106 2004-09-20 13:09:20Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifndef MPLOC_z2_CXX_H_
#define MPLOC_z2_CXX_H_

// It *is* possible to do something generic with the _field interface,
// since much of it is shared ; however, things would not be a direct
// transaltion from C to C++, so we stick to a dumb approach.
// For the record, one could make the different instances of the field
// class inherit from the same generic parent class. The benefit would be
// shared code for most of the ctor / dtor stuff.

#ifndef	CTORLOG
#define	CTORLOG(x)	x
#endif

class Mploc_z2_field {
    public:
	mplf_t		f;

	inline const mplf_t cval() const { return f; }
	
	inline Mploc_z2_field() : f(mplf_z2_create()) {}
	inline Mploc_z2_field(const mplf_t sf) : f(sf) {}

	inline Mploc_z2_field& operator=(const Mploc_z2_field& s) {
		/* This is to obey effc++ guidelines... ; in fact, we
		 * don't want this to be done, so yell
		 */
		std::cerr  << "Doing field=field !\n";
		f=mplf_z2_create();
		return *this;
	}
	inline ~Mploc_z2_field() {
		mplf_z2_destroy(&f);
	}
	inline Mploc_z2_field(const Mploc_z2_field &src) : f(mplf_z2_create()) {
		CTORLOG(std::cerr  << "CALLING COPY !\n");
	}
	static void inversion_configure(int m, int s) {} ;
};

class z2_elt {
    public:
	mploc_t	x;

	typedef Mploc_z2_field	field;

	//class Mploc_z2_field& f;	// a bit unfortunate.

	inline mploc_srcptr cval() const { return x; }
	inline mploc_ptr cval() { return x; }

	inline z2_elt(const mplf_t f) {
		mploc_z2_init(x, f);
	}
	inline z2_elt(const Mploc_z2_field& sf) {
		mploc_z2_init(x,sf.cval());
	}
	inline z2_elt(const z2_elt& src) {
		CTORLOG(std::cerr  << "CALLING COPY !\n");
		mploc_z2_init(x,src.cval()->f);
		mploc_z2_set(x,src.cval());
	}
	inline ~z2_elt() {
		mploc_z2_clear(x);
	}
	inline void set_reduce(z2_elt&  y) { set(y); }
	inline void reduce() { }
	inline int is_reduced() { return 1; }
	inline void normalize() { mploc_z2_normalize(x); }

	inline void force_precision(mp_size_t k) {
		mploc_z2_force_precision(x,k);
	}
	
	inline void set(const z2_elt& y, mp_size_t k = -1) {
		if (k==-1) k=y.cval()->prec;
		mploc_z2_set_change_precision(x,y.cval(),k);
	}
	inline void set(mp_limb_t u, mp_size_t k) {
		mploc_z2_set_ui(x,u,k);
	}
	inline void set(mpz_t z, mp_size_t k) {
		mploc_z2_set_mpz(x,z,k);
	}
	static inline void set(z2_elt&  z, const z2_elt&  y) {z.set(y);}

	inline void random(gmp_randstate_t prng, mp_size_t k) {
		mploc_z2_set_random(x,prng,k);
	}
	static inline void random(z2_elt&  z,gmp_randstate_t prng,mp_size_t k){
		z.random(prng,k);
	}
	inline void random(gmp_randstate_t prng, mp_size_t d, mp_size_t k) {
		random(prng,k);
	}
	static inline void random(z2_elt&  z, gmp_randstate_t prng,
			mp_size_t d, mp_size_t k) {
		z.random(prng,d,k);
	}

	inline int ok() const {
		return mploc_z2_is_good_looking(x);
	}

	inline int valuation() const {
		return mploc_z2_valuation(x);
	}

	inline int operator==(const z2_elt& y) const {
		return mploc_z2_eq(x,y.cval());
	}
	inline int operator==(mp_limb_t y) const {
		return mploc_z2_eq_ui(x,y);
	}
	/* Arithmetic operations are handled by static member functions.
	 * It's preferrable to a false illusion of a well-thought
	 * interface (don't have time for this :-))
	 */
	static inline void add(z2_elt& z, const z2_elt& x, const z2_elt& y) {
		mploc_z2_add(z.cval(),x.cval(),y.cval());
	}
	static inline void add(z2_elt& z, const z2_elt& x, mp_limb_t y) {
		mploc_z2_add_ui(z.cval(),x.cval(),y);
	}

	static inline void sub(z2_elt& z, const z2_elt& x, const z2_elt& y) {
		mploc_z2_sub(z.cval(),x.cval(),y.cval());
	}
	static inline void sub(z2_elt& z, const z2_elt& x, mp_limb_t y) {
		mploc_z2_sub_ui(z.cval(),x.cval(),y);
	}

	static inline void neg(z2_elt& x, const z2_elt& y) {
		mploc_z2_neg(x.cval(),y.cval());
	}

	static inline void mul(z2_elt& z, const z2_elt& x, const z2_elt& y) {
		mploc_z2_mul(z.cval(),x.cval(),y.cval());
	}
	static inline void sqr(z2_elt&  z, const z2_elt&  x) {
		mploc_z2_sqr(z.cval(),x.cval());
	}
	static inline void inv(z2_elt& x, const z2_elt& y) {
		mploc_z2_inv(x.cval(),y.cval());
	}
	static inline int invsqrt(z2_elt& x, const z2_elt& y) {
		return mploc_z2_invsqrt(x.cval(),y.cval());
	}
	static inline int sqrt(z2_elt& x, const z2_elt& y) {
		return mploc_z2_sqrt(x.cval(),y.cval());
	}
	static inline void lshift(z2_elt& z, const z2_elt& x, mp_size_t k) {
		mploc_z2_lshift(z.cval(),x.cval(),k);
	}
	static inline void rshift(z2_elt& z, const z2_elt& x, mp_size_t k) {
		mploc_z2_rshift(z.cval(),x.cval(),k);
	}
};

inline std::ostream& operator<< (std::ostream& o, z2_elt& x) {
	/* FIXME : Finish with an o.write ! */
	mploc_z2_print(x.cval());
	return o;
}

class z2_poly {
    public:
	mploc_z2x_t	x;

	typedef Mploc_z2_field	field;	// coeff.field actually...

	//class Mploc_z2_field& f;	// a bit unfortunate.

	inline mploc_z2x_srcptr cval() const { return x; }
	inline mploc_z2x_ptr cval() { return x; }

	inline z2_poly(mplf_t f) {
		CTORLOG(std::cerr<<"ctor\n");
		mploc_z2x_init(x, f);
	}
	inline z2_poly(const Mploc_z2_field& sf) {
		CTORLOG(std::cerr<<"ctor\n");
		mploc_z2x_init(x, sf.cval());
	}
	inline z2_poly(mploc_z2x_srcptr s) {
		CTORLOG(std::cerr<<"ctor\n");
		mploc_z2x_init(x, s->f);
		mploc_z2x_set(x,s);
	}
	inline z2_poly(z2_poly& src) {
		CTORLOG(std::cerr  << "CALLING COPY !\n");
		mploc_z2x_init(x,src.cval()->f);
		mploc_z2x_set(x,src.cval());
	}
	inline ~z2_poly() {
		CTORLOG(std::cerr<<"dtor\n");
		mploc_z2x_clear(x);
	}
	inline void normalize() { mploc_z2x_normalize(x); }

	inline void force_precision(mp_size_t k) {
		mploc_z2x_force_precision(x,k);
	}
	
	inline void set(z2_poly& y, mp_size_t k = -1) {
		if (k==-1) k=y.x->prec;
		mploc_z2x_set_change_precision(x,y.cval(),k);
	}
	inline void set(mp_limb_t u, mp_size_t k) {
		mploc_z2x_set_ui(x,u,k);
	}
	inline void set(mpz_t z, mp_size_t k) {
		mploc_z2x_set_mpz(x,z,k);
	}
	static inline void set(z2_poly&  z, z2_poly&  y) {z.set(y);}
	inline void random(gmp_randstate_t prng, mp_size_t d, mp_size_t k) {
		mploc_z2x_set_random(x,prng,d,k);
	}
	static inline void random(z2_poly&  z, gmp_randstate_t prng,
			mp_size_t d, mp_size_t k) {
		z.random(prng,d,k);
	}
	inline int ok() const {
		return mploc_z2x_is_good_looking(x);
	}

	inline int valuation() const {
		return mploc_z2x_valuation(x);
	}

	inline int operator==(const z2_poly& y) const {
		return mploc_z2x_eq(x,y.cval());
	}
	inline int operator==(mp_limb_t y) const {
		return mploc_z2x_eq_ui(x,y);
	}
	/* Arithmetic operations are handled by static member functions.
	 * It's preferrable to a false illusion of a well-thought
	 * interface (don't have time for this :-))
	 */
	static inline void add(z2_poly& z, const z2_poly& x, const z2_poly& y) {
		mploc_z2x_add(z.cval(),x.cval(),y.cval());
	}
	static inline void add(z2_poly& z, const z2_poly& x, mp_limb_t y) {
		mploc_z2x_add_ui(z.cval(),x.cval(),y);
	}

	static inline void sub(z2_poly& z, const z2_poly& x, const z2_poly& y) {
		mploc_z2x_sub(z.cval(),x.cval(),y.cval());
	}
	static inline void sub(z2_poly& z, const z2_poly& x, mp_limb_t y) {
		mploc_z2x_sub_ui(z.cval(),x.cval(),y);
	}

	static inline void neg(z2_poly& x, const z2_poly& y) {
		mploc_z2x_neg(x.cval(),y.cval());
	}

	static inline void mul(z2_poly& z, const z2_poly& x, const z2_poly& y) {
		mploc_z2x_mul(z.cval(),x.cval(),y.cval());
	}
	static inline void sqr(z2_poly& z, const z2_poly& x) {
		mploc_z2x_sqr(z.cval(),x.cval());
	}
	/* No inversion and square root, obviously... */
	static inline void lshift(z2_poly& z, const z2_poly& x, mp_size_t k) {
		mploc_z2x_lshift(z.cval(),x.cval(),k);
	}
	static inline void rshift(z2_poly& z, const z2_poly& x, mp_size_t k) {
		mploc_z2x_rshift(z.cval(),x.cval(),k);
	}
	/* Polynomials have also a handful of additional functions */
	inline mp_size_t deg() {
		return x->deg;
	}
	inline void realloc(int dir, mp_size_t d, mp_size_t k) {
		mploc_z2x_realloc(x,dir,d,k);
	}
	inline void set_xi(mp_size_t i, mp_size_t k) {
		mploc_z2x_set_xi(x,i,k);
	}
	inline void set_coeff(mp_size_t i, z2_elt& a) {
		mploc_z2x_set_coeff(x,i,a.cval());
	}
	inline void get_coeff(z2_elt&  a, mp_size_t i) {
		mploc_z2x_get_coeff(a.cval(),x,i);
	}
	static inline void mul(z2_poly& z, const z2_poly& x, const z2_elt& a) {
		mploc_z2x_mul_z2(z.cval(), x.cval(), a.cval());
	}
	static inline void mul_xi(z2_poly& z, const z2_poly& x, mp_size_t i) {
		mploc_z2x_mul_xi(z.cval(),x.cval(),i);
	}
	static inline void div_q_xi(z2_poly& z, const z2_poly& x, mp_size_t i) {
		mploc_z2x_div_q_xi(z.cval(),x.cval(),i);
	}
	static inline void div_r_xi(z2_poly& z, const z2_poly& x, mp_size_t i) {
		mploc_z2x_div_r_xi(z.cval(),x.cval(),i);
	}
	static inline void div(z2_poly& q, z2_poly& r, const z2_poly& a, const z2_poly& b) {
		mploc_z2x_div(q.cval(),r.cval(),a.cval(),b.cval());
	}
	int can_divide() {
		if (deg()==-1) return 0;
		z2_elt lc(x->coeff_field);
		get_coeff(lc,deg());
		return lc.valuation()==0;
	}
};

inline std::ostream& operator<< (std::ostream& o, z2_poly& x) {
	/* FIXME : Finish with an o.write ! */
	mploc_z2x_print(x.cval());
	return o;
}

#endif	/* MPLOC_z2x_CXX_H_ */
