/* $Id: zp_cxx.hpp 106 2004-09-20 13:09:20Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifndef MPLOC_ZP_CXX_H_
#define MPLOC_ZP_CXX_H_

// It *is* possible to do something generic with the _field interface,
// since much of it is shared ; however, things would not be a direct
// transaltion from C to C++, so we stick to a dumb approach.
// For the record, one could make the different instances of the field
// class inherit from the same generic parent class. The benefit would be
// shared code for most of the ctor / dtor stuff.

#ifndef	CTORLOG
#define	CTORLOG(x)	x
#endif

class Mploc_zp_field {
    public:
	mplf_t		f;

	inline const mplf_t cval() const { return f; }
	
	inline Mploc_zp_field(mpz_srcptr p) : f(mplf_zp_create(p)) {}
	inline Mploc_zp_field(const mplf_t sf) : f(sf) {}
	inline ~Mploc_zp_field() {
		mplf_zp_destroy(&f);
	}
	inline Mploc_zp_field(const Mploc_zp_field &src) :
		f(mplf_zp_create(src.cval()->p))
	{
		CTORLOG(std::cerr  << "CALLING COPY !\n");
	}
	inline Mploc_zp_field& operator=(const Mploc_zp_field& s)
	{
		std::cerr  << "Doing field=field !\n";
		f=mplf_zp_create(s.cval()->p);
		return *this;
	}

	static void inversion_configure(int m, int s) {
		mploc_zp_inversion_configure(m,s);
	}
};

class zp_elt {
    public:
	mploc_t	x;

	typedef Mploc_zp_field	field;

	//class Mploc_zp_field& f;	// a bit unfortunate.

	inline mploc_srcptr cval() const { return x; }
	inline mploc_ptr cval() { return x; }

	inline zp_elt(mplf_t f) {
		mploc_zp_init(x, f);
	}
	inline zp_elt(class Mploc_zp_field& sf) {
		mploc_zp_init(x,sf.cval());
	}
	inline zp_elt(zp_elt& src) {
		CTORLOG(std::cerr  << "CALLING COPY !\n");
		mploc_zp_init(x,src.cval()->f);
		mploc_zp_set(x,src.cval());
	}
	inline ~zp_elt() {
		mploc_zp_clear(x);
	}
	inline void set_reduce(zp_elt&  y) { set(y); }
	inline void reduce() { }
	inline int is_reduced() { return 1; }
	inline void normalize() { mploc_zp_normalize(x); }

	inline void force_precision(mp_size_t k) {
		mploc_zp_force_precision(x,k);
	}
	
	inline void set(const zp_elt&  y, mp_size_t k = -1) {
		if (k==-1) k=y.x->prec;
		mploc_zp_set_change_precision(x,y.cval(),k);
	}
	inline void set(mp_limb_t u, mp_size_t k) {
		mploc_zp_set_ui(x,u,k);
	}
	inline void set(mpz_t z, mp_size_t k) {
		mploc_zp_set_mpz(x,z,k);
	}
	static inline void set(zp_elt&  z, const zp_elt&  y) {z.set(y);}

	inline void random(gmp_randstate_t prng, mp_size_t k) {
		mploc_zp_set_random(x,prng,k);
	}
	static inline void random(zp_elt&  z,gmp_randstate_t prng,mp_size_t k){
		z.random(prng,k);
	}
	inline void random(gmp_randstate_t prng, mp_size_t d, mp_size_t k) {
		random(prng,k);
	}
	static inline void random(zp_elt&  z, gmp_randstate_t prng,
			mp_size_t d, mp_size_t k) {
		z.random(prng,d,k);
	}

	inline int ok() const {
		return mploc_zp_is_good_looking(x);
	}

	inline int valuation() const {
		return mploc_zp_valuation(x);
	}

	inline int operator==(const zp_elt&  y) const {
		return mploc_zp_eq(x,y.cval());
	}
	inline int operator==(mp_limb_t y) const {
		return mploc_zp_eq_ui(x,y);
	}
	/* Arithmetic operations are handled by static member functions.
	 * It's preferrable to a false illusion of a well-thought
	 * interface (don't have time for this :-))
	 */
	static inline void add(zp_elt&  z, const zp_elt&  x, const zp_elt&  y) {
		mploc_zp_add(z.cval(),x.cval(),y.cval());
	}
	static inline void add(zp_elt&  z, const zp_elt&  x, mp_limb_t y) {
		mploc_zp_add_ui(z.cval(),x.cval(),y);
	}

	static inline void sub(zp_elt&  z, const zp_elt&  x, const zp_elt&  y) {
		mploc_zp_sub(z.cval(),x.cval(),y.cval());
	}
	static inline void sub(zp_elt&  z, const zp_elt&  x, mp_limb_t y) {
		mploc_zp_sub_ui(z.cval(),x.cval(),y);
	}

	static inline void neg(zp_elt&  x, const zp_elt&  y) {
		mploc_zp_neg(x.cval(),y.cval());
	}

	static inline void mul(zp_elt&  z, const zp_elt&  x, const zp_elt&  y) {
		mploc_zp_mul(z.cval(),x.cval(),y.cval());
	}
	static inline void sqr(zp_elt&  z, const zp_elt&  x) {
		mploc_zp_sqr(z.cval(),x.cval());
	}
	static inline void inv(zp_elt&  x, const zp_elt&  y) {
		mploc_zp_inv(x.cval(),y.cval());
	}
	static inline int invsqrt(zp_elt&  x, const zp_elt&  y) {
		return mploc_zp_invsqrt(x.cval(),y.cval());
	}
	static inline int sqrt(zp_elt&  x, const zp_elt&  y) {
		return mploc_zp_sqrt(x.cval(),y.cval());
	}
	static inline void lshift(zp_elt&  z, const zp_elt&  x, mp_size_t k) {
		mploc_zp_lshift(z.cval(),x.cval(),k);
	}
	static inline void rshift(zp_elt&  z, const zp_elt&  x, mp_size_t k) {
		mploc_zp_rshift(z.cval(),x.cval(),k);
	}
};

inline std::ostream& operator<< (std::ostream& o, const zp_elt&  x) {
	/* FIXME : Finish with an o.write ! */
	mploc_zp_print(x.cval());
	return o;
}

class zp_poly {
    public:
	mploc_zpx_t	x;

	typedef Mploc_zp_field	field;	// coeff.field actually...
	
	//class Mploc_zp_field& f;	// a bit unfortunate.

	inline mploc_zpx_srcptr cval() const { return x; }
	inline mploc_zpx_ptr cval() { return x; }

	inline zp_poly(mplf_t f) {
		CTORLOG(std::cerr<<"ctor\n");
		mploc_zpx_init(x, f);
	}
	inline zp_poly(class Mploc_zp_field& sf) {
		CTORLOG(std::cerr<<"ctor\n");
		mploc_zpx_init(x, sf.cval());
	}
	inline zp_poly(mploc_zpx_srcptr s) {
		CTORLOG(std::cerr<<"ctor\n");
		mploc_zpx_init(x, s->f);
		mploc_zpx_set(x,s);
	}
	inline zp_poly(const zp_poly& src) {
		CTORLOG(std::cerr  << "CALLING COPY !\n");
		mploc_zpx_init(x,src.cval()->f);
		mploc_zpx_set(x,src.cval());
	}
	inline ~zp_poly() {
		CTORLOG(std::cerr<<"dtor\n");
		mploc_zpx_clear(x);
	}
	inline void normalize() { mploc_zpx_normalize(x); }

	inline void force_precision(mp_size_t k) {
		mploc_zpx_force_precision(x,k);
	}
	
	inline void set(const zp_poly& y, mp_size_t k = -1) {
		if (k==-1) k=y.x->prec;
		mploc_zpx_set_change_precision(x,y.cval(),k);
	}
	inline void set(mp_limb_t u, mp_size_t k) {
		mploc_zpx_set_ui(x,u,k);
	}
	inline void set(mpz_t z, mp_size_t k) {
		mploc_zpx_set_mpz(x,z,k);
	}
	static inline void set(zp_poly&  z, const zp_poly&  y) {z.set(y);}
	inline void random(gmp_randstate_t prng, mp_size_t d, mp_size_t k) {
		mploc_zpx_set_random(x,prng,d,k);
	}
	static inline void random(zp_poly&  z, gmp_randstate_t prng,
			mp_size_t d, mp_size_t k) {
		z.random(prng,d,k);
	}
	inline int ok() const {
		return mploc_zpx_is_good_looking(x);
	}

	inline int valuation() const {
		return mploc_zpx_valuation(x);
	}

	inline int operator==(const zp_poly& y) const {
		return mploc_zpx_eq(x,y.cval());
	}
	inline int operator==(mp_limb_t y) const {
		return mploc_zpx_eq_ui(x,y);
	}
	/* Arithmetic operations are handled by static member functions.
	 * It's preferrable to a false illusion of a well-thought
	 * interface (don't have time for this :-))
	 */
	static inline void add(zp_poly& z, const zp_poly& x, const zp_poly& y) {
		mploc_zpx_add(z.cval(),x.cval(),y.cval());
	}
	static inline void add(zp_poly& z, const zp_poly& x, mp_limb_t y) {
		mploc_zpx_add_ui(z.cval(),x.cval(),y);
	}

	static inline void sub(zp_poly& z, const zp_poly& x, const zp_poly& y) {
		mploc_zpx_sub(z.cval(),x.cval(),y.cval());
	}
	static inline void sub(zp_poly& z, const zp_poly& x, mp_limb_t y) {
		mploc_zpx_sub_ui(z.cval(),x.cval(),y);
	}

	static inline void neg(zp_poly& x, const zp_poly& y) {
		mploc_zpx_neg(x.cval(),y.cval());
	}

	static inline void mul(zp_poly& z, const zp_poly& x, const zp_poly& y) {
		mploc_zpx_mul(z.cval(),x.cval(),y.cval());
	}
	static inline void sqr(zp_poly& z, const zp_poly& x) {
		mploc_zpx_sqr(z.cval(),x.cval());
	}
	/* No inversion and square root, obviously... */
	static inline void lshift(zp_poly& z, const zp_poly& x, mp_size_t k) {
		mploc_zpx_lshift(z.cval(),x.cval(),k);
	}
	static inline void rshift(zp_poly& z, const zp_poly& x, mp_size_t k) {
		mploc_zpx_rshift(z.cval(),x.cval(),k);
	}
	/* Polynomials have also a handful of additional functions */
	inline mp_size_t deg() {
		return x->deg;
	}
	inline void realloc(int dir, mp_size_t d, mp_size_t k) {
		mploc_zpx_realloc(x,dir,d,k);
	}
	inline void set_xi(mp_size_t i, mp_size_t k) {
		mploc_zpx_set_xi(x,i,k);
	}
	inline void set_coeff(mp_size_t i, const zp_elt& a) {
		mploc_zpx_set_coeff(x,i,a.cval());
	}
	inline void get_coeff(zp_elt&  a, mp_size_t i) {
		mploc_zpx_get_coeff(a.cval(),x,i);
	}
	static inline void mul(zp_poly& z, const zp_poly& x, const zp_elt& a) {
		mploc_zpx_mul_zp(z.cval(), x.cval(), a.cval());
	}
	static inline void mul_xi(zp_poly& z, const zp_poly& x, mp_size_t i) {
		mploc_zpx_mul_xi(z.cval(),x.cval(),i);
	}
	static inline void div_q_xi(zp_poly& z, const zp_poly& x, mp_size_t i) {
		mploc_zpx_div_q_xi(z.cval(),x.cval(),i);
	}
	static inline void div_r_xi(zp_poly& z, const zp_poly& x, mp_size_t i) {
		mploc_zpx_div_r_xi(z.cval(),x.cval(),i);
	}
	static inline void div(zp_poly& q, zp_poly& r, const zp_poly& a, const zp_poly& b) {
		mploc_zpx_div(q.cval(),r.cval(),a.cval(),b.cval());
	}
	int can_divide() {
		if (deg()==-1) return 0;
		zp_elt lc(x->coeff_field);
		get_coeff(lc,deg());
		return lc.valuation()==0;
	}
};

inline std::ostream& operator<< (std::ostream& o, const zp_poly& x) {
	/* FIXME : Finish with an o.write ! */
	mploc_zpx_print(x.cval());
	return o;
}

#endif	/* MPLOC_ZP_CXX_H_ */
