/* $Id: zpn_cxx.hpp 106 2004-09-20 13:09:20Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifndef MPLOC_ZPN_CXX_H_
#define MPLOC_ZPN_CXX_H_

// It *is* possible to do something generic with the _field interface,
// since much of it is shared ; however, things would not be a direct
// translation from C to C++, so we stick to a dumb approach.
// For the record, one could make the different instances of the field
// class inherit from the same generic parent class. The benefit would be
// shared code for most of the ctor / dtor stuff.

#ifndef	CTORLOG
#define	CTORLOG(x)	x
#endif

class Mploc_zpn_field {
    public:
	mplf_t		f;

	inline const mplf_t cval() const { return f; }
	
	inline Mploc_zpn_field(const zp_poly& p):f(mplf_zpn_create(p.cval())) {}
	inline Mploc_zpn_field(const mplf_t sf) : f(sf) {}
	inline ~Mploc_zpn_field() {
		mplf_zpn_destroy(&f);
	}

	inline Mploc_zpn_field(const Mploc_zpn_field &src) :
		f(mplf_zpn_create(src.def_poly().cval()))
	{
		CTORLOG(std::cerr  << "CALLING COPY !\n");
	}

	inline Mploc_zpn_field& operator=(const Mploc_zpn_field& s)
	{
		std::cerr  << "Doing field=field !\n";
		f=mplf_zpn_create(s.def_poly().cval());
		return *this;
	}

	inline zp_poly& def_poly() const {
		return * new zp_poly(mplf_zpn_def_poly(f));
	}
	inline void set_reducer(zpx_reducer_t r, int t) {
		mplf_zpn_set_reducer(f,r,t);
	}
	static void inversion_configure(int m, int s) {} ;
};

class zpn_elt {
    public:
	mploc_zpx_t	x;

	typedef Mploc_zpn_field	field;
	
	//class Mploc_zpn_field& f;	// a bit unfortunate.

	inline mploc_zpx_srcptr cval() const { return x; }
	inline mploc_zpx_ptr cval() { return x; }

	inline zpn_elt(const mplf_t f) {
		mploc_zpn_init(x, f);
	}
	inline zpn_elt(const Mploc_zpn_field& sf) {
		mploc_zpn_init(x,sf.cval());
	}
	inline zpn_elt(const zpn_elt& src) {
		CTORLOG(std::cerr  << "CALLING COPY !\n");
		mploc_zpn_init(x,src.cval()->f);
		mploc_zpn_set(x,src.cval());
	}
	inline ~zpn_elt() {
		mploc_zpn_clear(x);
	}
	inline void set_reduce(const zpn_elt& y) {mploc_zpn_reduce(x,y.cval()); }
	inline void reduce() { set_reduce(*this); }
	inline int is_reduced() { return mploc_zpn_is_reduced(x); }
	inline void normalize() { mploc_zpn_normalize(x); }

	inline void force_precision(mp_size_t k) {
		mploc_zpn_force_precision(x,k);
	}
	
	inline void set(const zpn_elt& y, mp_size_t k = -1) {
		if (k==-1) k=y.x->prec;
		mploc_zpn_set_change_precision(x,y.cval(),k);
	}
	inline void set(mp_limb_t u, mp_size_t k) {
		mploc_zpn_set_ui(x,u,k);
	}
	inline void set(mpz_t z, mp_size_t k) {
		mploc_zpn_set_mpz(x,z,k);
	}
	static inline void set(zpn_elt&  z, const zpn_elt&  y) {z.set(y);}

	inline void random(gmp_randstate_t prng, mp_size_t k) {
		mploc_zpn_set_random(x,prng,k);
	}
	static inline void random(zpn_elt&  z,gmp_randstate_t prng,mp_size_t k){
		z.random(prng,k);
	}
	inline void random(gmp_randstate_t prng, mp_size_t d, mp_size_t k) {
		random(prng,k);
	}
	static inline void random(zpn_elt&  z, gmp_randstate_t prng,
			mp_size_t d, mp_size_t k) {
		z.random(prng,d,k);
	}

	inline int ok() const {
		return mploc_zpn_is_good_looking(x);
	}

	inline int valuation() const {
		return mploc_zpn_valuation(x);
	}

	inline int operator==(const zpn_elt& y) const {
		return mploc_zpn_eq(x,y.cval());
	}
	inline int operator==(mp_limb_t y) const {
		return mploc_zpn_eq_ui(x,y);
	}
	/* Arithmetic operations are handled by static member functions.
	 * It's preferrable to a false illusion of a well-thought
	 * interface (don't have time for this :-))
	 */
	static inline void add(zpn_elt& z, const zpn_elt& x, const zpn_elt& y) {
		mploc_zpn_add(z.cval(),x.cval(),y.cval());
	}
	static inline void add(zpn_elt& z, const zpn_elt& x, mp_limb_t y) {
		mploc_zpn_add_ui(z.cval(),x.cval(),y);
	}

	static inline void sub(zpn_elt& z, const zpn_elt& x, const zpn_elt& y) {
		mploc_zpn_sub(z.cval(),x.cval(),y.cval());
	}
	static inline void sub(zpn_elt& z, const zpn_elt& x, mp_limb_t y) {
		mploc_zpn_sub_ui(z.cval(),x.cval(),y);
	}

	static inline void neg(zpn_elt& x, const zpn_elt& y) {
		mploc_zpn_neg(x.cval(),y.cval());
	}

	static inline void mul(zpn_elt& z, const zpn_elt& x, const zpn_elt& y) {
		mploc_zpn_mul(z.cval(),x.cval(),y.cval());
	}
	static inline void sqr(zpn_elt&  z, const zpn_elt&  x) {
		mploc_zpn_sqr(z.cval(),x.cval());
	}
	static inline void inv(zpn_elt& x, const zpn_elt& y) {
		mploc_zpn_inv(x.cval(),y.cval());
	}
	static inline int invsqrt(zpn_elt& x, const zpn_elt& y) {
		return mploc_zpn_invsqrt(x.cval(),y.cval());
	}
	static inline int sqrt(zpn_elt& x, const zpn_elt& y) {
		return mploc_zpn_sqrt(x.cval(),y.cval());
	}
	static inline void lshift(zpn_elt& z, const zpn_elt& x, mp_size_t k) {
		mploc_zpn_lshift(z.cval(),x.cval(),k);
	}
	static inline void rshift(zpn_elt& z, const zpn_elt& x, mp_size_t k) {
		mploc_zpn_rshift(z.cval(),x.cval(),k);
	}
};

inline std::ostream& operator<< (std::ostream& o, const zpn_elt& x) {
	/* FIXME : Finish with an o.write ! */
	mploc_zpn_print(x.cval());
	return o;
}

#endif	/* MPLOC_ZPN_CXX_H_ */
