/* $Id: field_common.c 101 2004-09-17 15:28:03Z /CN=thome/emailAddress=thome@loria.fr $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gmp.h>
#include <assert.h>

#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include "gmp_hacks.h"
#include "field_def.h"
#include "field_common.h"

void generic_field_set(mp_limb_t * a,
		const mp_limb_t * b,
		const struct field * f)
{
	memcpy(a,b,f->size*sizeof(mp_limb_t));
}

int generic_field_eq(mp_limb_t * a,
		mp_limb_t * b,
		const struct field * f)
{
	f->reduce(a,f);
	f->reduce(b,f);
	return mpn_cmp(a,b,f->size)==0;
}

void generic_field_destroy(struct field * f)
{
	if (f->modulus)	free(f->modulus);
	if (f->minpoly)	free(f->minpoly);
	free(f);
}

void generic_field_exp(mp_limb_t * a,
		const mp_limb_t * r,
		const mp_limb_t * x,
		const mp_size_t n,
		const struct field * f)
{
	mp_limb_t * u;
	mp_size_t i,j,lead;	/* it is a signed type */
	mp_limb_t mask;

	for(i=n-1;i>=0 && x[i]==0;i--);
	if (i<0) {
		f->set_int(a,1,f);
		return;
	}

	j=GMP_LIMB_BITS-1;
	mask=(1UL<<j);

	for(;(x[i]&mask)==0;j--,mask>>=1);

	lead=i*GMP_LIMB_BITS+j;	/* Ensured. */

	f->set(a,r,f);

	u=malloc(f->size*sizeof(mp_limb_t));
	for(;lead>0;lead--) {
		if (j--==0) {
			i--;
			j=GMP_LIMB_BITS-1;
			mask=(1UL<<j);
		} else {
			mask>>=1;
		}
	       	if (x[i]&mask) {
			f->mul(u,a,a,f);
			f->mul(a,u,r,f);
		} else {
			f->mul(u,a,a,f);
			f->set(a,u,f);
		}
	}
	free(u);
}
