/* $Id: rf_layer_gf2n_ntl.cpp 101 2004-09-17 15:28:03Z /CN=thome/emailAddress=thome@loria.fr $ */

#undef __DEPRECATED

#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include <gmp.h>

#include <NTL/GF2X.h>
#include <NTL/GF2XFactoring.h>
#include <NTL/GF2E.h>
#include <NTL/GF2EX.h>
#include <NTL/GF2EXFactoring.h>

#include <cassert>

extern "C" {
#include "mploc.h"
#include "manu.h"

#include "z2x.h"
#include "z2n.h"
}

#include "rf_layer_gf2n_ntl.hpp"

NTL_USE_NNS

void ntl_attach_GF2E(mplf_t f)
{
	GF2X p;
	p=0;
	mploc_z2x_srcptr pptr=mplf_z2n_def_poly(f);
	for(int i=0;i<=pptr->deg;i++) {
		SetCoeff(p,i,pptr->c[i*pptr->cw]);
	}

	GF2E::init(p);
}

void ntl_detach_GF2E(mplf_t f)
{
	/*
	GF2X p;
	p=0;
	GF2E::init(p);
	*/
	// We prefer to do nothing here. The above fails badly, as NTL
	// checks pretty early that it's been given something sane (and 0
	// isn't).
}

rf_elt new_rfe(mplf_t f UNUSED_VARIABLE)
{
	return (void*) new GF2E();
}

void _delete_rfe(rf_elt * a, mplf_t f UNUSED_VARIABLE)
{
	delete (GF2E *) *a;
	*a=NULL;
}

void rfe_set_mploc_coeff(rf_elt xbar, mploc_z2x_srcptr x, mp_size_t k, mplf_t f)
{
	GF2X p;
	p=0;
	ASSERT(k<x->prec);
	mp_size_t wi=k/GMP_LIMB_BITS, wp=k%GMP_LIMB_BITS;

	for(int i=0;i<=x->deg;i++) {
		SetCoeff(p,i,x->c[i*x->cw + wi]>>wp);
	}
	*(GF2E*)xbar=to_GF2E(p);
}

void rfe_set_mploc(rf_elt xbar, mploc_z2x_srcptr x, mplf_t f)
{
	rfe_set_mploc_coeff(xbar,x,0,f);
}

void mploc_set_rfe(mploc_z2x_ptr x, rf_elt a, mplf_t f)
{
	const GF2X& p=rep(*(GF2E*)a);
	mploc_z2x_clobber(x);
	mploc_z2x_realloc(x,1,f->degree-1,1);
	x->deg=f->degree-1;
	for(int i=0;i<f->degree;i++) {
		x->c[i*x->cw]=rep(coeff(p,i));
	}
	mploc_z2x_normalize(x);
}

/* This one is more subtle than mploc_set_rfe */
void mploc_coeff_set_rfe(mploc_z2x_ptr x, mp_size_t k, rf_elt a, mplf_t f)
{
	const GF2X& p=rep(*(GF2E*)a);
	ASSERT(k<x->prec);
	mp_size_t wi=k/GMP_LIMB_BITS, wp=k%GMP_LIMB_BITS;

	/* Make sure there's room ! */
	mploc_z2x_realloc(x,1,f->degree-1,x->prec);

	x->deg=f->degree-1;
	for(int i=0;i<f->degree;i++) {
		x->c[i*x->cw + wi] &= ~(1UL<<wp);
		x->c[i*x->cw + wi] |= (rep(coeff(p,i))&1UL)<<wp;
	}
	mploc_z2x_normalize(x);
}

#if 0
/* Useful for debugging */
#include <iostream>
#include <string>

static std::ostream& operator<<(std::ostream& o, const rf_elt& pp)
{
	int k=0;
	const GF2E& p=*(GF2E*)pp;
	//std::ostringstream o;
	for(int i=0;i<=deg(rep(p));i++) {
		if (coeff(rep(p),i)==0) continue;
		if (k) o << " + ";
		if (i==0) o << "1";
		else if (i==1) o << "X";
		else o << "X^" << i;
		k++;
	}
	if (!k) o << "0";
	return o;
}
#endif

void RF_sqrt(rf_elt y,rf_elt x, mplf_t f)
{
	*(GF2E*)y=*(GF2E*)x;
	for(int i=0;i<f->degree-1;i++) {
		*(GF2E*)y=sqr(*(GF2E*)y);
	}
}

void RF_inv(rf_elt y,rf_elt x, mplf_t f)
{
	inv(*(GF2E*)y,*(GF2E*)x);
}

void RF_invsqrt(rf_elt y,rf_elt x, mplf_t f)
{
	RF_inv(y,x,f);
	RF_sqrt(y,y,f);
}

int RF_is_zero(rf_elt x, mplf_t f)
{
	return *(GF2E*)x == 0;
}

void RF_mul(rf_elt z, rf_elt y,rf_elt x, mplf_t f)
{
	mul(*(GF2E*)z, *(GF2E*)x, *(GF2E*)y);
}

/* return x such that x^p-x=a ; 
 * such a root exists iff tr(a)==0 ; return 0 if this is not the
 * case.
 */
   
int RF_artin_schreier_root(rf_elt x,rf_elt a,mplf_t f)
{
	GF2EX p;
	p=0;
	SetCoeff(p,2,1);
	SetCoeff(p,1,-1);
	SetCoeff(p,0,-*(GF2E*)a);
	FindRoot(*(GF2E*)x,p);
	return 1;
}


/* tells whether the trace of a is zero */
int RF_has_trace_zero(rf_elt a,mplf_t f)
{
	return trace(*(GF2E*)a)==0L;
}

