/* $Id: z2n.h 204 2005-01-20 08:12:15Z /CN=gaudry/emailAddress=gaudry@polytechnique.fr $ */

#ifndef Z2N_H_
#define Z2N_H_

#ifdef __cplusplus
extern "C" {
#endif

/* Most functions in this interface are macros for the polynomial
 * analogue.
 */

extern mplf_t mplf_z2n_create(mploc_z2x_srcptr);
extern void mplf_z2n_destroy(mplf_t * pf);

extern mploc_z2x_srcptr mplf_z2n_def_poly(mplf_t);

struct z2x_reducer_t {
	unsigned long magic;
	void (*reduce)(mploc_z2n_ptr, mploc_z2n_srcptr);
	int handles_overlaps;
	void (*hook)(mplf_t, struct z2x_reducer_t *);
	void (*unhook)(mplf_t);
	void * priv;
};

/* A generic reduction-capable function takes the destination and the
 * source.
 *
 * The defining polynomial, if needed, can be accessed from the reducing
 * function using mplf_z2n_def_poly
 */

/* third arg tells whether the function handles overlaps. NULL means :
 * turn back to default. */
extern void mplf_z2n_set_reducer(mplf_t, struct z2x_reducer_t *);

extern void mplf_z2n_set_reducer_preinv(mplf_t);

extern void mploc_z2n_reduce(mploc_z2n_ptr, mploc_z2n_srcptr);
#define mploc_z2n_is_reduced(x) ((x)->deg<((x)->f->degree))

/************************************************************************/

extern void mploc_z2n_init(mploc_z2n_ptr, mplf_t);
extern void mploc_z2n_clear(mploc_z2n_ptr);

#define mploc_z2n_force_precision(a, b)	mploc_z2x_force_precision(a, b)
#define mploc_z2n_set(a, b)		mploc_z2x_set(a, b)
#define mploc_z2n_set_ui(a, b, c)	mploc_z2x_set_ui(a, b, c)
#define mploc_z2n_set_mpz(a, b, c)	mploc_z2x_set_mpz(a, b, c)
#define mploc_z2n_set_random(a, b, c)	mploc_z2x_set_random(a, b, a->f->degree-1, c)
#define mploc_z2n_add(a, b, c)		mploc_z2x_add(a, b, c)
#define mploc_z2n_add_ui(a, b, c)	mploc_z2x_add_ui(a, b, c)
#define mploc_z2n_sub(a, b, c)		mploc_z2x_sub(a, b, c)
#define mploc_z2n_sub_ui(a, b, c)	mploc_z2x_sub_ui(a, b, c)
#define mploc_z2n_normalize(a)		mploc_z2x_normalize(a)

#define mploc_z2n_neg(a,b)		mploc_z2x_neg(a,b)


#define mploc_z2n_set_change_precision(a,b,c)	mploc_z2x_set_change_precision(a,b,c) 

extern void mploc_z2n_mul(mploc_z2n_ptr, mploc_z2n_srcptr, mploc_z2n_srcptr);
extern void mploc_z2n_inv(mploc_z2n_ptr, mploc_z2n_srcptr);
extern int mploc_z2n_invsqrt(mploc_z2n_ptr, mploc_z2n_srcptr);
extern void mploc_z2n_sqr(mploc_z2n_ptr, mploc_z2n_srcptr);
extern int mploc_z2n_sqrt(mploc_z2n_ptr, mploc_z2n_srcptr);

#define mploc_z2n_print(a)		mploc_z2x_print(a)
#define mploc_z2n_valuation(a)		mploc_z2x_valuation(a)

#define mploc_z2n_is_good_looking(a)	mploc_z2x_is_good_looking(a)
#define mploc_z2n_lshift(z,x,n)		mploc_z2x_lshift(z,x,n)
#define mploc_z2n_lshift_quick(z,x,n)   mploc_z2x_lshift_quick(z,x,n)

extern void mploc_z2n_rshift(mploc_z2n_ptr, mploc_z2n_srcptr, mp_size_t);

extern int mploc_z2n_eq(mploc_z2n_srcptr, mploc_z2n_srcptr);
extern int mploc_z2n_eq_ui(mploc_z2n_srcptr, mp_limb_t);

#define mploc_z2n_is_zero(x)		(mploc_z2n_eq_ui((x),0))
#define mploc_z2n_is_one(x)		(mploc_z2n_eq_ui((x),1))

extern void mploc_z2n_trace(mploc_ptr, mploc_z2n_srcptr);


#ifdef __cplusplus
}
#endif

#endif	/* Z2N_H_ */
