/* $Id: zpn.h 91 2004-09-16 15:02:07Z /CN=thome/emailAddress=thome@loria.fr $ */

#ifndef ZPN_H_
#define ZPN_H_

#ifdef __cplusplus
extern "C" {
#endif

/* Most functions in this interface are macros for the polynomial
 * analogue.
 */

extern mplf_t mplf_zpn_create(mploc_zpx_srcptr);
extern void mplf_zpn_destroy(mplf_t * pf);

extern mploc_zpx_srcptr mplf_zpn_def_poly(mplf_t);

/* A generic reduction-capable function takes the destination and the
 * source.
 *
 * The defining polynomial, if needed, can be accessed from the reducing
 * function using mplf_zpn_def_poly
 */
typedef void (*zpx_reducer_t)(mploc_zpn_ptr, mploc_zpn_srcptr);

/* third arg tells whether the function handles overlaps. NULL means :
 * turn back to default. */
extern void mplf_zpn_set_reducer(mplf_t, zpx_reducer_t, int);

extern void mploc_zpn_reduce(mploc_zpn_ptr, mploc_zpn_srcptr);
#define mploc_zpn_is_reduced(x) ((x)->deg<((x)->f->degree))

/************************************************************************/

extern void mploc_zpn_init(mploc_zpn_ptr, mplf_t);
extern void mploc_zpn_clear(mploc_zpn_ptr);

#define mploc_zpn_force_precision(a, b)	mploc_zpx_force_precision(a, b)
#define mploc_zpn_set(a, b)		mploc_zpx_set(a, b)
#define mploc_zpn_set_ui(a, b, c)	mploc_zpx_set_ui(a, b, c)
#define mploc_zpn_set_mpz(a, b, c)	mploc_zpx_set_mpz(a, b, c)
#define mploc_zpn_set_random(a, b, c)	mploc_zpx_set_random(a, b, a->f->degree-1, c)
#define mploc_zpn_add(a, b, c)		mploc_zpx_add(a, b, c)
#define mploc_zpn_add_ui(a, b, c)	mploc_zpx_add_ui(a, b, c)
#define mploc_zpn_sub(a, b, c)		mploc_zpx_sub(a, b, c)
#define mploc_zpn_sub_ui(a, b, c)	mploc_zpx_sub_ui(a, b, c)
#define mploc_zpn_normalize(a)		mploc_zpx_normalize(a)

#define mploc_zpn_neg(a,b)		mploc_zpx_neg(a,b)


#define mploc_zpn_set_change_precision(a,b,c)	mploc_zpx_set_change_precision(a,b,c) 

extern void mploc_zpn_mul(mploc_zpn_ptr, mploc_zpn_srcptr, mploc_zpn_srcptr);
extern void mploc_zpn_inv(mploc_zpn_ptr, mploc_zpn_srcptr);
extern int mploc_zpn_invsqrt(mploc_zpn_ptr, mploc_zpn_srcptr);
extern int mploc_zpn_sqrt(mploc_zpn_ptr, mploc_zpn_srcptr);
extern void mploc_zpn_sqr(mploc_zpn_ptr, mploc_zpn_srcptr);

#define mploc_zpn_print(a)		mploc_zpx_print(a)
#define mploc_zpn_valuation(a)		mploc_zpx_valuation(a)

#define mploc_zpn_is_good_looking(a)	mploc_zpx_is_good_looking(a)
#define mploc_zpn_lshift(z,x,n)		mploc_zpx_lshift(z,x,n)

extern void mploc_zpn_rshift(mploc_zpn_ptr, mploc_zpn_srcptr, mp_size_t);

extern int mploc_zpn_eq(mploc_zpn_srcptr, mploc_zpn_srcptr);
extern int mploc_zpn_eq_ui(mploc_zpn_srcptr, mp_limb_t);

#define mploc_zpn_is_zero(x)		(mploc_zpn_eq_ui((x),0))
#define mploc_zpn_is_one(x)		(mploc_zpn_eq_ui((x),1))


#ifdef __cplusplus
}
#endif

#endif	/* ZPN_H_ */
