/* $Id: zpx.c 186 2004-10-13 11:55:26Z /CN=thome/emailAddress=thome@loria.fr $ */

/***************************************************************************/
/* HEADERS */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* This is all the GMP stuff I need... */
#include <gmp.h>

#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include <assert.h>

#include "gmp_hacks.h"
#include "mploc.h"
#include "zp.h"
#include "zpx.h"

#include "manu.h"

#define BAD_CODE_WARNING				\
	BEGIN_BLOCK					\
		static int i;				\
		if (!i++) {				\
			WARNING("This is bad code");	\
		}					\
	END_BLOCK


/***************************************************************************/
/* ENTRY POINTS -- POLYNOMIAL INTERFACE */

/* Initialize a polynomial (initialized with degree -1, precision 0, no
 * allocated coeffs.
 *
 * Note that f is mostly useless, but we use it for population count on
 * init and close (one poly counts as 1)
 */
void mploc_zpx_init(mploc_zpx_ptr x, mplf_t f)
{
	x->prec=0;
	x->deg=-1;
	x->alloc_coeffs=0;
	x->c = NULL;
	x->f = NULL;
	x->coeff_field=f;
#ifdef DEBUG_MPLOC
	x->coeff_field->people++;
#endif
}

/* frees up the poly */
void mploc_zpx_clear(mploc_zpx_ptr x)
{
	int i;
#ifdef DEBUG_MPLOC
	x->coeff_field->people--;
#endif
	for(i=0;i<x->alloc_coeffs;i++) {
		mpz_clear(x->c[i]);
	}
	free(x->c);
	memset(x,0,sizeof(mploc_zpx_ptr));
}

int mploc_zpx_is_good_looking_maybe_nonnormal(mploc_zpx_srcptr x)
{
	int i;
	if (x->alloc_coeffs < x->deg + 1) {
		fprintf(stderr, "Poly has unallocated coeffs\n");
		return 0;
	}
	if (x->deg<-1) {
		fprintf(stderr, "Poly has unexpected degree (<-1)\n");
		return 0;
	}
	for(i=0;i<=x->deg;i++) {
		if (SIZ(x->c[i])<0) {
			fprintf(stderr, "coeff is negative integer\n");
			return 0;
		}

		if (ALLOC(x->c[i]) < SIZ(x->c[i])) {
			fprintf(stderr, "coeff w/ insufficient allocation\n");
			return 0;
		}
	}
	if (x->coeff_field == NULL) {
		fprintf(stderr, "Poly has no parent field\n");
		return 0;
	}
#ifdef	DEBUG_MPLOC
	if (x->coeff_field->people <= 0) {
		fprintf(stderr, "Poly has parent field with <=0 elts.\n");
		return 0;
	}
#endif

	return 1;
}

int mploc_zpx_is_good_looking(mploc_zpx_srcptr x)
{
	if (!mploc_zpx_is_good_looking_maybe_nonnormal(x)) {
		return 0;
	}
	if (x->prec==0 && x->deg!=-1) {
		fprintf(stderr, "Poly has precision 0 and degree>=0\n");
		return 0;
	}
	/* Maybe check for zero leading coeff */
	return 1;
}

/* Reallocation.
 *
 * Most comments present in z2x obviously apply here too.
 */

void mploc_zpx_clobber(mploc_zpx_ptr x)
{
	x->deg=-1;
	x->prec=0;
}

int mploc_zpx_realloc(mploc_zpx_ptr x,
		int dir,
		mp_size_t deg,
		mp_size_t prec)
{
	int i,res=0;

	ASSERT(dir==-1 || dir==1);
	ASSERT(deg>=-1);
	ASSERT(prec>=0);

	/* We require an explicit clobber before a shrink that goes below
	 * the current degree. See z2x.c for a discussion about this. */
	ASSERT(deg>=x->deg || x->deg==-1);
	// Hence all MIN(deg,x->deg) have been replaced by x->deg

	if (dir>0 && x->alloc_coeffs < (deg+1)) {
		x->c = realloc(x->c, (deg+1)*sizeof(mpz_t));
		for(i=x->alloc_coeffs;i<=deg;i++) {
			mpz_init(x->c[i]);
		}
		x->alloc_coeffs=deg+1;
		res=1;
	} else if (dir<0 && x->alloc_coeffs > (deg+1)) {
		for(i=deg+1;i<x->alloc_coeffs;i++) {
			mpz_clear(x->c[i]);
		}
		x->c = realloc(x->c, (deg+1)*sizeof(mpz_t));
		x->alloc_coeffs=deg+1;
		res=1;
	}

	x->prec=prec;
	/* x->deg *IS NOT CHANGED* here ; see z2x.c for a discussion on
	 * the rationale of this. */

#if 0	/* made impossible by the assert above. */
	if (x->deg > deg)
		x->deg=deg;
#endif

	ASSERT(mploc_zpx_is_good_looking_maybe_nonnormal(x));
	return res;
}

/* This is a pretty trivial application of the _realloc() interface() */

/* Change the precision, extending if necessary. See discussion above.
 * x->prec *is changed* by this operation.
 */
void mploc_zpx_force_precision(mploc_zpx_ptr x, mp_size_t k)
{
	if (k<x->prec) {
		mploc_zpx_realloc(x,1,x->deg,k);
		mploc_zpx_normalize(x);
	} else {
		mploc_zpx_realloc(x,1,x->deg,k);
	}
}

/* Remove the leading zero coefficients in x */
void mploc_zpx_normalize(mploc_zpx_ptr x)
{
	int i;

	for(i=x->deg;i>=0;i--) {
		mpz_mod(x->c[i],x->c[i],p_to_the(x->prec,x->coeff_field));
		if (mpz_cmp_ui(x->c[i],0)!=0) break;
	}
	x->deg=i;
	for(;i>=0;i--) {
		mpz_mod(x->c[i],x->c[i],p_to_the(x->prec,x->coeff_field));
	}

	ASSERT(mploc_zpx_is_good_looking(x));
}

/* There is *no* _insert_zero procedure ; because it's not useful.  */
/* x <- y + O(2^k) */
void mploc_zpx_set_change_precision(mploc_zpx_ptr x, mploc_zpx_srcptr y, mp_size_t k)
{
	int i;

	if (EXPECT_FALSE(x==y)) {
		mploc_zpx_force_precision(x,k);
		/* result is normalized above */
		return;
	}

	ASSERT(x!=y);

	mploc_zpx_clobber(x);
	mploc_zpx_realloc(x,1,y->deg,k);
	x->deg=y->deg;

	for(i=0;i<=y->deg;i++) {
		mpz_set(x->c[i],y->c[i]);
	}

	mploc_zpx_normalize(x);

	ASSERT(mploc_zpx_is_good_looking(x));
}

/* x <- y */
void mploc_zpx_set(mploc_zpx_ptr x, mploc_zpx_srcptr y)
{
	mploc_zpx_set_change_precision(x,y,y->prec);
}

/* x <- X^i + O(2^prec) */
void mploc_zpx_set_xi(mploc_zpx_ptr x, mp_size_t i, mp_size_t prec)
{
	int j;

	mploc_zpx_clobber(x);
	mploc_zpx_realloc(x,1,i,prec);
	x->deg=i;

	for(j=0;j<x->deg;j++) mpz_set_ui(x->c[j],0);
	mpz_set_ui(x->c[x->deg],1);

	ASSERT(mploc_zpx_is_good_looking(x));
}

/* [X^i] x <- u + O(2^prec)
 *
 * If gprec is the precision of other elements, gprec <- MIN(prec,gprec)
 */
void mploc_zpx_set_coeff(mploc_zpx_ptr x, mp_size_t i, mploc_srcptr a)
{
	int j;
	mploc_zpx_realloc(x,1,MAX(x->deg,i),MIN(x->prec,a->prec));
	for(j=x->deg+1;j<i;j++) mpz_set_ui(x->c[j],0);
	x->deg=MAX(x->deg,i);
	mpz_set(x->c[x->deg],a->x);

	/* A whole bunch of funny things can happen... Better to
	 * normalize here. */
	mploc_zpx_normalize(x);

	ASSERT(mploc_zpx_is_good_looking(x));
}

/* u <- [X^i] x */
void mploc_zpx_get_coeff(mploc_ptr a, mploc_zpx_srcptr x, mp_size_t i)
{
	ASSERT(i>=0);

	mploc_zp_force_precision(a,x->prec);
	if (i<=x->deg)
		mpz_set(a->x,x->c[i]);
	else
		mpz_set_ui(a->x,0);

	ASSERT(mploc_zp_is_good_looking(a));
}

/* x <- u + O(2^prec) */
void mploc_zpx_set_ui(mploc_zpx_ptr x, mp_limb_t u, mp_size_t prec)
{
	if (EXPECT_FALSE(!u || prec==0)) {
		mploc_zpx_clobber(x);
		mploc_zpx_realloc(x,1,-1,prec);
		x->deg=-1;
		ASSERT(mploc_zpx_is_good_looking(x));
		return;
	}

	mploc_zpx_clobber(x);
	mploc_zpx_realloc(x,1,0,prec);
	x->deg=0;
	mpz_set_ui(x->c[0],u);

	/* If p^prec divides u, we have 0 ; a division is necessary to
	 * detect it...
	 */
	mploc_zpx_normalize(x);

	ASSERT(mploc_zpx_is_good_looking(x));
}

/* x <- a + O(2^prec) */
void mploc_zpx_set_mpz(mploc_zpx_ptr x, mpz_srcptr a, mp_size_t prec)
{
	if (!SIZ(a)) {
		mploc_zpx_clobber(x);
		mploc_zpx_realloc(x,1,-1,prec);
		x->deg=-1;
		return;
	}

	mploc_zpx_clobber(x);
	mploc_zpx_realloc(x,1,0,prec);
	x->deg=0;

	if (SIZ(a)>0) {
		mpz_set(x->c[0],a);
	} else {
		mpz_add(x->c[0],p_to_the(prec,x->coeff_field),a);
	}
	
	/* same business */
	mploc_zpx_normalize(x);

	ASSERT(mploc_zpx_is_good_looking(x));
}

/* x <- Random poly of degree d and precision k
 *
 * There's a bit of code duplication here. (GMP's fault)
 */
void mploc_zpx_set_random(mploc_zpx_ptr x, gmp_randstate_t prng,
		mp_size_t d, mp_size_t prec)
{
	int i;

	mploc_zpx_clobber(x);
	mploc_zpx_realloc(x,1,d,prec);
	x->deg=d;

	for(i=0;i<=x->deg;i++) {
		mpz_urandomm(x->c[i],prng,p_to_the(prec,x->coeff_field));
	}
	mploc_zpx_normalize(x);

	ASSERT(mploc_zpx_is_good_looking(x));
}

/* z <- x + y */
void mploc_zpx_add(mploc_zpx_ptr z, mploc_zpx_srcptr x, mploc_zpx_srcptr y)
{
	mp_size_t ad;
	mp_size_t i;

	ad = MAX(x->deg,y->deg);
	if (z!=x && z!=y) mploc_zpx_clobber(z);
	mploc_zpx_realloc(z,1,ad,MIN(x->prec,y->prec));
	/* DO NOT change z->deg to ad now (in case z==x or z==y) */
	
	for(i=0;i<=MIN(x->deg,y->deg);i++) {
		mpz_add(z->c[i],x->c[i],y->c[i]);
	}
	if (x->deg > y->deg) {
		for(;i<=ad;i++) {
			mpz_set(z->c[i],x->c[i]);
		}
	} else if (x->deg < y->deg) {
		for(;i<=ad;i++) {
			mpz_set(z->c[i],y->c[i]);
		}
	}

	/* Do the degree change here in case z=x and deg(y)>deg(x) !  */
	z->deg = ad;

	/* Again, precision issues might kill us :-( */
	mploc_zpx_normalize(z);

	ASSERT(mploc_zpx_is_good_looking(z));
}

/* Add the integer u to constant coefficient */
void mploc_zpx_add_ui(mploc_zpx_ptr z, mploc_zpx_srcptr x, mp_limb_t u)
{
	if (z!=x) mploc_zpx_clobber(z);
	mploc_zpx_realloc(z,1,MAX(0,x->deg),x->prec);
	mploc_zpx_set(z,x);
	z->deg=MAX(x->deg,0);
	mpz_add_ui(z->c[0],z->c[0],u);

	if (x->deg==0)
		mploc_zpx_normalize(z);
	else
		mpz_mod(z->c[0],z->c[0],p_to_the(z->prec,x->coeff_field));

	ASSERT(mploc_zpx_is_good_looking(z));
}

/* z <- x - y */
void mploc_zpx_sub(mploc_zpx_ptr z, mploc_zpx_srcptr x, mploc_zpx_srcptr y)
{
	mp_size_t ad;
	mp_size_t i;

	ad = MAX(x->deg,y->deg);
	if (z!=x && z!=y) mploc_zpx_clobber(z);
	mploc_zpx_realloc(z,1,ad,MIN(x->prec,y->prec));
	/* DO NOT change z->deg to ad now (in case z==x or z==y) */
	
	for(i=0;i<=MIN(x->deg,y->deg);i++) {
		mpz_sub(z->c[i],x->c[i],y->c[i]);
	}
	if (x->deg > y->deg) {
		for(;i<=ad;i++) {
			mpz_set(z->c[i],x->c[i]);
		}
	} else if (x->deg < y->deg) {
		for(;i<=ad;i++) {
			mpz_neg(z->c[i],y->c[i]);
			/* This is OK as long as a normalization is
			 * performed at the end */
		}
	}

	/* Do the degree change here in case z=x and deg(y)>deg(x) !  */
	z->deg = ad;

	/* Again, precision issues might kill us :-( */
	mploc_zpx_normalize(z);

	ASSERT(mploc_zpx_is_good_looking(z));
}

/* Subtract integer u to the constant coeff */
void mploc_zpx_sub_ui(mploc_zpx_ptr z, mploc_zpx_srcptr x, unsigned long u)
{
	if (z!=x) mploc_zpx_clobber(z);
	mploc_zpx_realloc(z,1,MAX(0,x->deg),x->prec);
	mploc_zpx_set(z,x);
	z->deg=MAX(x->deg,0);
	mpz_sub_ui(z->c[0],z->c[0],u);

	if (x->deg==0)
		mploc_zpx_normalize(z);
	else
		mpz_mod(z->c[0],z->c[0],p_to_the(z->prec,x->coeff_field));

	ASSERT(mploc_zpx_is_good_looking(z));
}

void mploc_zpx_neg(mploc_zpx_ptr z, mploc_zpx_srcptr x)
{
	mp_size_t i;

	if (z!=x) mploc_zpx_clobber(z);
	mploc_zpx_realloc(z,1,x->deg,x->prec);
	z->deg=x->deg;

	for(i=0;i<=z->deg;i++) {
		mpz_sub(z->c[i],p_to_the(x->prec,x->coeff_field),x->c[i]);
	}
	ASSERT(mploc_zpx_is_good_looking(z));
}


/* z <- X^i * x */
void mploc_zpx_mul_xi(mploc_zpx_ptr z, mploc_zpx_srcptr x, mp_size_t i)
{
	int j;

	if (z!=x) mploc_zpx_clobber(z);
	mploc_zpx_realloc(z,1,x->deg+i,x->prec);
	/* z->deg changed later */

	if (i>=0) {
		for(j=x->deg;j>=0;j--)
			mpz_set(z->c[j+i],x->c[j]);
		for(j=0;j<i;j++)
			mpz_set_ui(z->c[j],0);
	} else {
		for(j=-i;j<=x->deg;j++)
			mpz_set(z->c[j+i],x->c[j]);
	}

	z->deg = x->deg + i;

	ASSERT(mploc_zpx_is_good_looking(z));
}

/* z <- x div X^i */
void mploc_zpx_div_q_xi(mploc_zpx_ptr z, mploc_zpx_srcptr x, mp_size_t i)
{
	if (i>x->deg) {
		/* The precision is x->prec because the whole polynomial
		 * is understood modulo 2^prec */
		mploc_zpx_set_ui(z,0,x->prec);
	} else {
		mploc_zpx_mul_xi(z,x,-i);
	}
}

/* z -> x mod X^i */
void mploc_zpx_div_r_xi(mploc_zpx_ptr z, mploc_zpx_srcptr x, mp_size_t i)
{
	mploc_zpx_set(z,x);
	if (i>x->deg)
		return;

	z->deg=i-1;
	
	mploc_zpx_normalize(z);

	ASSERT(mploc_zpx_is_good_looking(z));
}

/* z -> a * x */
void mploc_zpx_mul_zp(mploc_zpx_ptr z, mploc_zpx_srcptr x, mploc_srcptr a)
{
	int i=0;

	if (z!=x) mploc_zpx_clobber(z);
	mploc_zpx_realloc(z,1,x->deg,MIN(x->prec,a->prec));
	z->deg=x->deg;

	/* Perhaps one could do ``short'' products in some way (unclear) */
	/* Then code should be shared with mploc_zp_mul */
	BAD_CODE_WARNING;

	for(i=0;i<=x->deg;i++) {
		mpz_mul(z->c[i],x->c[i],a->x);
	}

	mploc_zpx_normalize(z);
	
	ASSERT(mploc_zpx_is_good_looking(z));
}

/* prints x to stdout */
void mploc_zpx_print(mploc_zpx_srcptr x)
{
	mp_size_t i,j UNUSED_VARIABLE;
	int some=0;

	for(i=0;i<=x->deg;i++) {
		if (!SIZ(x->c[i])) { continue; }
		if (some) printf(" + ");
		some=1;
		switch(i) {
			case 0:
				gmp_printf("%Zd",x->c[i]); break;
			case 1:
				if (mpz_cmp_ui(x->c[i],1)!=0) {
					gmp_printf("%Zd*",x->c[i]);
				}
				printf("X");
				break;
			default:
				if (mpz_cmp_ui(x->c[i],1)!=0) {
					gmp_printf("%Zd*",x->c[i]);
				}
				printf("X^%lu",i);
				break;
		}
	}

	if (some) printf(" + ");
	printf("O(p^%d)", (int) x->prec);
}

int mploc_zpx_valuation(mploc_zpx_srcptr x)
{
	mpz_t 	q,r;
	int	v,k;
	int	i;

	v=x->prec;
	mpz_init(q);
	mpz_init(r);
	for(i=0;v && i<=x->deg;i++) {
		mpz_set(q,x->c[i]);
		for(k=0;k<v;k++) {
			mpz_fdiv_qr(q,r,q,x->coeff_field->p);
			if (mpz_cmp_ui(r,0)!=0)
				break;
		}
		v = MIN(k,v);
	}
	mpz_clear(r);
	mpz_clear(q);
	return v;
}
		
int mploc_zpx_eq(mploc_zpx_srcptr x, mploc_zpx_srcptr y)
{
	mploc_zpx_t z;
	int r;

	mploc_zpx_init(z,x->coeff_field);
	mploc_zpx_sub(z,x,y);
	r=z->deg==-1;
	mploc_zpx_clear(z);

	return r;
}

int mploc_zpx_eq_ui(mploc_zpx_srcptr x, mp_limb_t v)
{
	mploc_zpx_t z;
	int r;

	if (x->deg)
		return 0;

	mploc_zpx_init(z,x->coeff_field);
	mploc_zpx_sub_ui(z,x,v);
	r=z->deg==-1;
	mploc_zpx_clear(z);

	return r;
}
	
void mploc_zpx_lshift(mploc_zpx_ptr z, mploc_zpx_srcptr x, mp_size_t n)
{
	mp_size_t i;

	if (EXPECT_FALSE(n==0)) {
		mploc_zpx_set(z,x);
	}

	if (z!=x) mploc_zpx_clobber(z);
	mploc_zpx_realloc(z,1,x->deg,n+x->prec);
	z->deg=x->deg;

	for(i=0;i<=x->deg;i++) {
		mpz_mul(z->c[i],x->c[i],p_to_the(n,x->coeff_field));
	}

	// Fortunately no normalization is necessary...

	ASSERT(mploc_zpx_is_good_looking(z));
}

void mploc_zpx_rshift(mploc_zpx_ptr z, mploc_zpx_srcptr x, mp_size_t n)
{
	mp_size_t i;
	
	if (EXPECT_FALSE(n==0)) {
		mploc_zpx_set(z,x);
	}

	if (z!=x) mploc_zpx_clobber(z);
	mploc_zpx_realloc(z,1,x->deg,x->prec);
	/* cannot require x->prec-n if z==x :-( */
	z->deg=x->deg;

	for(i=0;i<=x->deg;i++) {
		mpz_fdiv_q(z->c[i],x->c[i],p_to_the(n,x->coeff_field));
	}

	/* This is a bit crude... */
	z->prec-=n;

	/* Head coeff might have become 0 ; apart from this, coefficients
	 * need not be mod'ed ; The mods contained in the normalize step
	 * will therefore be single comparisons, and won't cost much.
	 * (And we expect no significant penalty compared to the overall
	 * cost of the rshift op here...)
	 */
	mploc_zpx_normalize(z);

	ASSERT(mploc_zpx_is_good_looking(z));
}

/****************************************************************************/
/* Multiplication of polynomials */

/* Low-level routines used */

static int zpx_kara_threshold = 32;
/* Should depend on SIZ(modulo) */
#define ZPX_KARA_MUL_THRESHOLD(modulo)	zpx_kara_threshold

void set_zpx_kara_threshold(int x) { zpx_kara_threshold = x; }

#if 0	/* This is the Z2 data */
/* This has been tuned for p3 celeron 700MHz, 128KB cache */
static const mp_size_t kara_thresholds[] = {
	/* cw == 1 */	12,
	/* cw == 2 */	8,
	/* cw == 3 */	6, 6, 6, 6, 6, 6, 6,
	/* cw == 10 */	4, 4, 4, 4, 4};

#define	zpx_KARA_MUL_THRESHOLD(cw)	(((cw)>=15)?2:kara_thresholds[(cw)-1])
#endif

#define EXCHPTR(a,na,b,nb)	BEGIN_BLOCK	\
			mpz_t *	t;	\
			mp_size_t	n;	\
			t=a;a=b;b=t;		\
			n=na;na=nb;nb=n;	\
		END_BLOCK

/* schoolbook */
static void zpx_mul_basecase(mpz_t * r, mpz_t * a, mpz_t * b, mpz_srcptr modulo, mp_size_t na, mp_size_t nb);

/* karatsuba */
static void zpx_kara_mul_n(mpz_t * r, mpz_t * a, mpz_t * b, mpz_srcptr modulo, mp_size_t n, mpz_t * ws);

static void zpx_mul_back(mpz_t * r, mpz_t * a, mpz_t * b, mpz_srcptr modulo, mp_size_t na, mp_size_t nb, mplf_t f);

#define	zpx_mul_n_wrap(r,a,b,modulo,d,ws)	((d)<ZPX_KARA_MUL_THRESHOLD(modulo))?zpx_mul_basecase(r,a,b,modulo,d,d):zpx_kara_mul_n(r,a,b,modulo,d,ws)

/* Of course we should implement a kara_square ; wins 2x if basecase
 * manages to win that much. */
void mploc_zpx_sqr(mploc_zpx_ptr z, mploc_zpx_srcptr x)
{
	BAD_CODE_WARNING;
	mploc_zpx_mul(z,x,x);
}

void mploc_zpx_mul(mploc_zpx_ptr z, mploc_zpx_srcptr x, mploc_zpx_srcptr y)
{
	mploc_zpx_t target;
	int use_tmp;
	mp_size_t totaldeg;

	totaldeg=(x->deg==-1||y->deg==-1)?-1:(x->deg + y->deg);

	if (totaldeg==-1) {
		/* Do a ``soft clobber'', since we don't want to thrash
		 * the precision */
		z->deg=-1;
		/* also set the precision */
		mploc_zpx_realloc(z,1,-1,MIN(x->prec,y->prec));
		return;
	}

	/* If z is one of the operands, this saves some work. */
	if (z!=x && z!=y) mploc_zpx_clobber(z);
	mploc_zpx_realloc(z,1,totaldeg,MIN(x->prec,y->prec));
	/* Don't set z->deg now... */

	use_tmp =  (z->c == x->c || z->c == y->c);
	if (use_tmp) {
		mploc_zpx_init(target,x->coeff_field);
		mploc_zpx_realloc(target,1,totaldeg,MIN(x->prec,y->prec));
	}


	/* time to call the back end */
	zpx_mul_back(use_tmp?target->c:z->c,
			x->c,y->c,p_to_the(z->prec,z->coeff_field),
			x->deg+1,y->deg+1,z->coeff_field);

	/* NOW we can set z->deg */
	if (use_tmp) {
		target->deg = totaldeg;
	} else {
		z->deg = totaldeg;
	}

	if (use_tmp) {
		mploc_zpx_set(z,target);
		mploc_zpx_clear(target);
	}

	/* Dunno if it's needed. Probably... XXX */
	mploc_zpx_normalize(z);
}

/* Helpers for add and sub
 *
 * TODO: could actually be useful for mploc_zpx_add too.
 */
static inline void zpx_add(mpz_t * r,
		mpz_t * a,
		mpz_t * b,
		mp_size_t n)
{
	int i;
	for(i=0;i<n;i++) mpz_add(r[i],a[i],b[i]);
}

static inline void zpx_sub(mpz_t * r,
		mpz_t * a,
		mpz_t * b,
		mp_size_t n)
{
	int i;
	for(i=0;i<n;i++) mpz_sub(r[i],a[i],b[i]);
}

static inline void zpx_cpy(mpz_t * r,
		mpz_t * a,
		mp_size_t n)
{
	int i;
	for(i=0;i<n;i++) mpz_set(r[i],a[i]);
}

static inline void zpx_neg(mpz_t * r,
		mpz_t * a,
		mp_size_t n)
{
	int i;
	for(i=0;i<n;i++) mpz_neg(r[i],a[i]);
}

/*
 * This is the code for unbalanced mults. It does its best to do the good
 * number of symmetric multiplications.
 *
 * This follows mpn_mul from gmp-4.1.
 *
 * na and nb here are the number of coefficients involved (hence degree+1).
 *
 * na and nb multiplied form na+nb-1 coefficients (pain...)
 */
static void zpx_mul_back(mpz_t * r, mpz_t * a, mpz_t * b, mpz_srcptr modulo, mp_size_t na, mp_size_t nb, mplf_t f)
{
	mploc_zpx_t tmp, kara_tmp;
	mp_size_t l;

	if (nb > na)	EXCHPTR(a,na,b,nb);

	if (nb<ZPX_KARA_MUL_THRESHOLD(modulo)) {
		zpx_mul_basecase(r,a,b,modulo,na,nb);
		return;
	}

	/* OK, on va avoir besoin de faire du karatsuba, avec une taille
	 * au plus n. On s'alloue un buffer correspondant.
	 */

	mploc_zpx_init(tmp,f);
	mploc_zpx_init(kara_tmp,f);

	/* Precision doesn't matter so much ; tmp and kara_tmp are merely
	 * used as arrays of mpz_t's
	 */
	mploc_zpx_realloc(kara_tmp,1,(nb-1),0);

	zpx_kara_mul_n(r,a,b,modulo,nb,kara_tmp->c);

	if (na!=nb) {
		r  += nb;
		na -= nb;
		a  += nb;
		l   = nb-1;

		/* l is the number of coefficients set in r by the last
		 * mult.  operation. Of course, since r might have been
		 * shifted, l takes this into account. Since we're
		 * swapping pointers back and forth, there's not much
		 * point in trying to ``guess'' what l is from the
		 * current values of na and nb (and it would be pretty
		 * hard, in fact).
		 */
		
		if (na < nb) EXCHPTR(a,na,b,nb);

		mploc_zpx_realloc(tmp,1,
				((nb>=ZPX_KARA_MUL_THRESHOLD(modulo)?nb:na)+nb-1)-1,
				0);
		
		while (nb>=ZPX_KARA_MUL_THRESHOLD(modulo)) {
			
			/* Loop invariant here: the product of a and b,
			 * with respectively na and nb coefficients, has
			 * to be stored into r, where l coefficents are
			 * already set.
			 */

			zpx_kara_mul_n(tmp->c,a,b,modulo,nb,kara_tmp->c);

			if (l<= (2*nb - 1)) {
				zpx_add(r,r,tmp->c,l);
				zpx_cpy(r+l,tmp->c+l,(2*nb-1-l));
				l=2*nb-1;
			} else {
				zpx_add(r,r,tmp->c,2*nb-1);
			}

			r  += nb;
			a  += nb;
			na -= nb;
			l  -= nb;
			
			if (na < nb) EXCHPTR(a,na,b,nb);
		}
		
		if (nb != 0) {
			zpx_mul_basecase(tmp->c,a,b,modulo,na,nb);
			if (l <= na + nb - 1) {
				zpx_add(r,r,tmp->c,l);
				zpx_cpy(r+l,tmp->c+l,na+nb-1-l);
			} else {
				zpx_add(r,r,tmp->c,na + nb-1);
			}
		}

	}

	mploc_zpx_clear(tmp);
	mploc_zpx_clear(kara_tmp);
}

/* karatsuba */
static void zpx_kara_mul_n(mpz_t * r, mpz_t * a, mpz_t * b, mpz_srcptr modulo, mp_size_t n, mpz_t * ws)
{
	/* ws doit avoir de la place pour n entiers */

	if (n&1)	{	// odd number of coeffs.
		mp_size_t p;

		p = n>>1;
		/* q = p + 1 */

		/* A = al + X^p * ah    deg al = p-1, deg ah = q-1
		 * B = bl + X^p * bh    deg bl = p-1, deg bh = q-1
		 *
		 * alpha = al - ah      deg alpha = q-1
		 * beta  = bh - bl      deg beta =  q-1
		 *
		 * gamma = alpha * beta + al*bl + ah * bh = al * bh + ah * bl
		 *
		 * A*B = al * bl + X^p * gamma + X^2p * ah * bh
		 */
		zpx_sub(r,a,a+p,p);
		zpx_neg(r+p,a+n-1,1);

		/* Now r[0...q[ is alpha = al-ah */

		zpx_sub(r+p+1,b+p,b,p);
		zpx_cpy(r+n,b+n-1,1);	// p+q=n XXX

		/* Now r[q...2q[ is beta = bh - bl */
		/* Up to n + 1 coefficients in r are used. */

		zpx_mul_n_wrap(ws,r,r+p+1,modulo,p+1,r+(n+1));
		zpx_mul_n_wrap(r+2*p,a+p,b+p,modulo,p+1,r);

		zpx_add(ws,ws,r+2*p,n);
		zpx_cpy(r+p,ws,p);
		zpx_add(r+2*p,r+2*p,ws+p,p+1);

		/* Now r[p...2n-1[ contains
		 * X^p*(alpha * beta + ah*bh) + X^2p*ah*bh
		 */

		zpx_mul_n_wrap(ws,a,b,modulo,p,r);

		zpx_cpy(r,ws,p);
		zpx_add(r+p,r+p,ws+p,p-1);
		zpx_add(r+p,r+p,ws,n-2);
	} else {	// Even n
		mp_size_t p;

		p = n>>1;	// and q = p

		zpx_sub(r,a,a+p,p);
		zpx_sub(r+p,b+p,b,p);

		zpx_mul_n_wrap(ws,r,r+p,modulo,p,r+n);
		zpx_mul_n_wrap(r+2*p,a+p,b+p,modulo,p,r);

		zpx_add(ws,ws,r+2*p,n-1);
		zpx_cpy(r+p,ws,p);
		zpx_add(r+2*p,r+2*p,ws+p,p-1);

		zpx_mul_n_wrap(ws,a,b,modulo,p,r);

		zpx_cpy(r,ws,p);
		zpx_add(r+p,r+p,ws+p,p-1);
		zpx_add(r+p,r+p,ws,n-1);
#if 0
		/* This is completely generic */
	} else {
		mp_size_t p, q, p;
		p = n>>1;
		q = n - p;
		zpx_sub(r,a,a+p,p);
		if (q!=p) zpx_neg(r+p,a+(n-1),1);
		zpx_sub(r+q,b+p,b,p);
		if (q!=p) zpx_cpy(r+n,b+(n-1),1);// p+q=n XXX
		zpx_mul_n_wrap(ws,r,r+q,modulo,q,r+2*q);
		zpx_mul_n_wrap(r+2*p,a+p,b+p,modulo,q,r);
		zpx_add(ws,ws,r+2*p,2*q-1);
		zpx_cpy(r+p,ws,p);
		zpx_add(r+2*p,r+2*p,ws+p,2*q-1-p);
		zpx_mul_n_wrap(ws,a,b,modulo,p,r);
		zpx_cpy(r,ws,p);
		zpx_add(r+p,r+p,ws+p,p-1);
		zpx_add(r+p,r+p,ws,2*p-1);
#endif
	}
}

/* schoolbook */
static void zpx_mul_basecase(mpz_t * r, mpz_t * a, mpz_t * b, mpz_srcptr modulo, mp_size_t na, mp_size_t nb)
{
	mp_size_t i,j;

	ASSERT(r!=a && r!=b);

	ASSERT(na && nb);

	for(i=0;i<na+nb-1;i++) mpz_set_ui(r[i],0);

	for(i=0;i<na;i++) {
		for(j=0;j<nb;j++) {
			mpz_addmul(r[i+j],a[i],b[j]);
		}
	}

	for(i=0;i<na+nb-1;i++) mpz_mod(r[i],r[i],modulo);
}

/* q or r might be unneeded. */
void mploc_zpx_div(mploc_zpx_ptr q, mploc_zpx_ptr r,
		mploc_zpx_srcptr a, mploc_zpx_srcptr b)
{
	mploc_zpx_t	ta;
	mp_size_t	np;
	mploc_t		c;
	mpz_t		m;	/* Used for multiplications */
	mp_size_t	j;

	if (b->deg==-1)	DIVIDE_BY_ZERO();

	np=MIN(a->prec,b->prec);

	ASSERT(q || r);	/* Otherwise user is an idiot */
			/* XXX Do remove if user *is* an idiot :-) */

	ASSERT(q!=b && r!= b);
	
	if (q) {
		if (q!=a && q!=b) mploc_zpx_clobber(q);
		mploc_zpx_realloc(q,1,MAX(a->deg-b->deg,q->deg),np);
	}
	if (r) {
		if (r!=a && r!=b) mploc_zpx_clobber(r);
		mploc_zpx_realloc(r,1,MAX(r->deg,b->deg-1),np);
	}
	if (q && r)
		ASSERT(q!=r);

	if (a->deg < b->deg) {
		/* Set r before, in case q==a ! */
		if (r) mploc_zpx_set_change_precision(r,a,np);
		if (q) q->deg=-1;
		if (q) ASSERT(mploc_zpx_is_good_looking(q));
		if (r) ASSERT(mploc_zpx_is_good_looking(r));
		return;
	}

	mploc_zpx_init(ta,a->coeff_field);
	mploc_zpx_set_change_precision(ta,a,np);

	mploc_zp_init(c,a->coeff_field);
	mpz_init(m);

	mploc_zpx_get_coeff(c,b,b->deg);
	mploc_zp_force_precision(c,np);
	mploc_zp_inv(c,c);

	if (q) {
		q->deg=a->deg-b->deg;
		for(j=0;j<q->alloc_coeffs;j++) mpz_set_ui(q->c[j],0);
	}

	if (r) r->deg=b->deg-1;

#define	i	(ta->deg)
	for(;i>=b->deg;) {
		mpz_mul(m,ta->c[i],c->x);
		mpz_mod(m,m,p_to_the(np,a->coeff_field));
		for(j=1;j<=b->deg;j++) {
			mpz_submul(ta->c[i-j],b->c[b->deg-j],m);
		}
		/* Here we try to bit slightly more subtle than the
		 * normalization step, as lazyness undoubtedly pas off on
		 * the low limbs.
		 */
		if (q) mpz_set(q->c[i-b->deg],m);
		for(;--i>=b->deg;) {
			mpz_mod(ta->c[i],ta->c[i],p_to_the(np,a->coeff_field));
			if (mpz_cmp_ui(ta->c[i],0)!=0)
				break;
		}
	}
#undef	i
	if (r) zpx_cpy(r->c,ta->c,b->deg);

	/* No need to normalize q */
	if (r) mploc_zpx_normalize(r);

	mpz_clear(m);
	mploc_zp_clear(c);
	mploc_zpx_clear(ta);

	if (q) ASSERT(mploc_zpx_is_good_looking(q));
	if (r) ASSERT(mploc_zpx_is_good_looking(r));
}
